import base64
from unittest.mock import ANY, create_autospec

from django.test import SimpleTestCase, override_settings

from anymail.exceptions import AnymailInsecureWebhookWarning
from anymail.signals import inbound, tracking

from .utils import AnymailTestMixin, ClientWithCsrfChecks


def event_handler(sender, event, esp_name, **kwargs):
    """Prototypical webhook signal handler"""
    pass


@override_settings(ANYMAIL={"WEBHOOK_SECRET": "username:password"})
class WebhookTestCase(AnymailTestMixin, SimpleTestCase):
    """Base for testing webhooks

    - connects webhook signal handlers
    - sets up basic auth by default (since most ESP webhooks warn if it's not enabled)
    """

    client_class = ClientWithCsrfChecks

    def setUp(self):
        super().setUp()
        # Use correct basic auth by default (individual tests can override):
        self.set_basic_auth()

        # Install mocked signal handlers
        self.tracking_handler = create_autospec(event_handler)
        tracking.connect(self.tracking_handler)
        self.addCleanup(tracking.disconnect, self.tracking_handler)

        self.inbound_handler = create_autospec(event_handler)
        inbound.connect(self.inbound_handler)
        self.addCleanup(inbound.disconnect, self.inbound_handler)

    def set_basic_auth(self, username="username", password="password"):
        """Set basic auth for all subsequent test client requests"""
        credentials = base64.b64encode(
            "{}:{}".format(username, password).encode("utf-8")
        ).decode("utf-8")
        self.client.defaults["HTTP_AUTHORIZATION"] = "Basic {}".format(credentials)

    def clear_basic_auth(self):
        self.client.defaults.pop("HTTP_AUTHORIZATION", None)

    def assert_handler_called_once_with(
        self, mockfn, *expected_args, **expected_kwargs
    ):
        """Verifies mockfn was called with expected_args and at least expected_kwargs.

        Ignores *additional* actual kwargs
        (which might be added by Django signal dispatch).
        (This differs from mock.assert_called_once_with.)

        Returns the actual kwargs.
        """
        self.assertEqual(mockfn.call_count, 1)
        actual_args, actual_kwargs = mockfn.call_args
        self.assertEqual(actual_args, expected_args)
        for key, expected_value in expected_kwargs.items():
            if expected_value is ANY:
                self.assertIn(key, actual_kwargs)
            else:
                self.assertEqual(actual_kwargs[key], expected_value)
        return actual_kwargs

    def get_kwargs(self, mockfn):
        """Return the kwargs passed to the most recent call to mockfn"""
        self.assertIsNotNone(mockfn.call_args)  # mockfn hasn't been called yet
        actual_args, actual_kwargs = mockfn.call_args
        return actual_kwargs


class WebhookBasicAuthTestCase(WebhookTestCase):
    """Common test cases for webhook basic authentication.

    Instantiate for each ESP's webhooks by:
    - subclassing
    - defining call_webhook to invoke the ESP's webhook
    - adding or overriding any tests as appropriate
    """

    def __init__(self, methodName="runTest"):
        if self.__class__ is WebhookBasicAuthTestCase:
            # don't run these tests on the abstract base implementation
            methodName = "runNoTestsInBaseClass"
        super().__init__(methodName)

    def runNoTestsInBaseClass(self):
        pass

    #: subclass set False if other webhook verification used
    should_warn_if_no_auth = True

    def call_webhook(self):
        # Concrete test cases should call a webhook via self.client.post,
        # and return the response
        raise NotImplementedError()

    @override_settings(ANYMAIL={})  # Clear the WEBHOOK_AUTH settings from superclass
    def test_warns_if_no_auth(self):
        if self.should_warn_if_no_auth:
            with self.assertWarns(AnymailInsecureWebhookWarning):
                response = self.call_webhook()
        else:
            with self.assertDoesNotWarn(AnymailInsecureWebhookWarning):
                response = self.call_webhook()
        self.assertEqual(response.status_code, 200)

    def test_verifies_basic_auth(self):
        response = self.call_webhook()
        self.assertEqual(response.status_code, 200)

    def test_verifies_bad_auth(self):
        self.set_basic_auth("baduser", "wrongpassword")
        response = self.call_webhook()
        self.assertEqual(response.status_code, 400)

    def test_verifies_missing_auth(self):
        self.clear_basic_auth()
        response = self.call_webhook()
        self.assertEqual(response.status_code, 400)

    @override_settings(ANYMAIL={"WEBHOOK_SECRET": ["cred1:pass1", "cred2:pass2"]})
    def test_supports_credential_rotation(self):
        """You can supply a list of basic auth credentials, and any is allowed"""
        self.set_basic_auth("cred1", "pass1")
        response = self.call_webhook()
        self.assertEqual(response.status_code, 200)

        self.set_basic_auth("cred2", "pass2")
        response = self.call_webhook()
        self.assertEqual(response.status_code, 200)

        self.set_basic_auth("baduser", "wrongpassword")
        response = self.call_webhook()
        self.assertEqual(response.status_code, 400)



#!/usr/bin/env python

# usage: python runtests.py [tests.test_x tests.test_y.SomeTestCase ...]

import os
import re
import sys
import warnings
from pathlib import Path

import django
from django.conf import settings
from django.test.utils import get_runner


def find_test_settings():
    """
    Return dotted path to Django settings compatible with current Django version.

    Finds highest tests.test_settings.settings_N_M.py where N.M is <= Django version.
    (Generally, default Django settings don't change meaningfully between Django
    releases, so this will fall back to the most recent settings when there isn't an
    exact match for the current version, while allowing creation of new settings
    files to test significant differences.)
    """
    django_version = django.VERSION[:2]  # (major, minor)
    found_version = None  # (major, minor)
    found_path = None

    for settings_path in Path("tests/test_settings").glob("settings_*.py"):
        try:
            (major, minor) = re.match(
                r"settings_(\d+)_(\d+)\.py", settings_path.name
            ).groups()
            settings_version = (int(major), int(minor))
        except (AttributeError, TypeError, ValueError):
            raise ValueError(
                f"File '{settings_path!s}' doesn't match settings_N_M.py"
            ) from None
        if settings_version <= django_version:
            if found_version is None or settings_version > found_version:
                found_version = settings_version
                found_path = settings_path

    if found_path is None:
        raise ValueError(f"No suitable test_settings for Django {django.__version__}")

    # Convert Path("test/test_settings/settings_N_M.py")
    # to dotted module "test.test_settings.settings_N_M":
    return ".".join(found_path.with_suffix("").parts)


def setup_and_run_tests(test_labels=None):
    """Discover and run project tests. Returns number of failures."""
    test_labels = test_labels or ["tests"]

    tags = envlist("ANYMAIL_ONLY_TEST")
    exclude_tags = envlist("ANYMAIL_SKIP_TESTS")

    # In automated testing, don't run live tests unless specifically requested
    if envbool("CONTINUOUS_INTEGRATION") and not envbool("ANYMAIL_RUN_LIVE_TESTS"):
        exclude_tags.append("live")

    if tags:
        print("Only running tests tagged: %r" % tags)
    if exclude_tags:
        print("Excluding tests tagged: %r" % exclude_tags)

    # show DeprecationWarning and other default-ignored warnings:
    warnings.simplefilter("default")

    settings_module = find_test_settings()
    print(f"Using settings module {settings_module!r}.")
    os.environ["DJANGO_SETTINGS_MODULE"] = settings_module
    django.setup()

    TestRunner = get_runner(settings)
    test_runner = TestRunner(verbosity=1, tags=tags, exclude_tags=exclude_tags)
    return test_runner.run_tests(test_labels)


def runtests(test_labels=None):
    """Run project tests and exit"""
    # Used as setup test_suite: must either exit or return a TestSuite
    failures = setup_and_run_tests(test_labels)
    sys.exit(bool(failures))


def envbool(var, default=False):
    """Returns value of environment variable var as a bool, or default if not set/empty.

    Converts `'true'` and similar string representations to `True`,
    and `'false'` and similar string representations to `False`.
    """
    # Adapted from the old :func:`~distutils.util.strtobool`
    val = os.getenv(var, "").strip().lower()
    if val == "":
        return default
    elif val in ("y", "yes", "t", "true", "on", "1"):
        return True
    elif val in ("n", "no", "f", "false", "off", "0"):
        return False
    else:
        raise ValueError("invalid boolean value env[%r]=%r" % (var, val))


def envlist(var):
    """Returns value of environment variable var split in a comma-separated list.

    Returns an empty list if variable is empty or not set.
    """
    val = [item.strip() for item in os.getenv(var, "").split(",")]
    if val == [""]:
        # "Splitting an empty string with a specified separator returns ['']"
        val = []
    return val


if __name__ == "__main__":
    runtests(test_labels=sys.argv[1:])



from urllib.parse import quote

from ..exceptions import AnymailAPIError
from ..message import AnymailRecipientStatus
from ..utils import get_anymail_setting
from .base_requests import AnymailRequestsBackend, RequestsPayload


class EmailBackend(AnymailRequestsBackend):
    """
    Scaleway Transactional Email API Backend
    """

    esp_name = "Scaleway"

    def __init__(self, **kwargs):
        """Init options from Django settings"""
        esp_name = self.esp_name
        self.secret_key = get_anymail_setting(
            "secret_key", esp_name=esp_name, kwargs=kwargs, allow_bare=True
        )
        self.project_id = get_anymail_setting(
            "project_id", esp_name=esp_name, kwargs=kwargs
        )
        region = get_anymail_setting(
            "region", esp_name=esp_name, kwargs=kwargs, default="fr-par"
        )
        api_url_template = get_anymail_setting(
            "api_url",
            esp_name=esp_name,
            kwargs=kwargs,
            default=(
                "https://api.scaleway.com/transactional-email/v1alpha1/regions/{region}/"
            ),
        )
        api_url = api_url_template.format(region=quote(region, safe=""))
        if not api_url.endswith("/"):
            api_url += "/"
        super().__init__(api_url, **kwargs)

    def build_message_payload(self, message, defaults):
        return ScalewayPayload(message, defaults, self)

    _recipient_status_map = {
        # Scaleway send status -> Anymail status.
        # (In practice, only "sending" seems to be reported.
        # Invalid addresses cause an API failure.
        # Blocked addresses show as "sending" and are rejected later.)
        "unknown": "unknown",
        "new": "queued",
        "sending": "queued",
        "sent": "sent",
        "failed": "failed",
        "canceled": "failed",
    }

    def parse_recipient_status(self, response, payload, message):
        parsed_response = self.deserialize_json_response(response, payload, message)
        statuses = {}
        try:
            emails = parsed_response["emails"]
        except (KeyError, TypeError):
            raise AnymailAPIError(
                "Invalid response from Scaleway API",
                email_message=message,
                payload=payload,
                response=response,
                backend=self,
            )

        for email_info in emails:
            recipient = email_info.get("mail_rcpt")
            message_id = email_info.get("id")
            status = email_info.get("status")
            anymail_status = AnymailRecipientStatus(
                message_id=message_id,
                status=self._recipient_status_map.get(status, "unknown"),
            )
            if recipient:
                statuses[recipient] = anymail_status
        return statuses


class ScalewayPayload(RequestsPayload):
    def __init__(self, message, defaults, backend, *args, **kwargs):
        self.project_id = backend.project_id
        http_headers = kwargs.pop("headers", {})
        http_headers["X-Auth-Token"] = backend.secret_key
        http_headers["Content-Type"] = "application/json"
        super().__init__(
            message, defaults, backend, headers=http_headers, *args, **kwargs
        )

    def get_api_endpoint(self):
        return "emails"

    def init_payload(self):
        self.data = {
            "project_id": self.project_id,
        }

    def _scaleway_email(self, email):
        """Expand an Anymail EmailAddress into Scaleway's {"email", "name"} dict"""
        result = {"email": email.addr_spec}
        if email.display_name:
            result["name"] = email.display_name
        return result

    def set_from_email(self, email):
        self.data["from"] = self._scaleway_email(email)

    def set_recipients(self, recipient_type, emails):
        assert recipient_type in {"to", "cc", "bcc"}
        if emails:
            self.data[recipient_type] = [
                self._scaleway_email(email) for email in emails
            ]

    def set_subject(self, subject):
        if subject:
            self.data["subject"] = subject

    def set_text_body(self, body):
        if body:
            self.data["text"] = body

    def set_html_body(self, html):
        if html:
            self.data["html"] = html

    def add_attachment(self, attachment):
        if attachment.inline:
            self.unsupported_feature("inline attachments")
        self.data.setdefault("attachments", []).append(
            {
                "name": attachment.name,
                "type": attachment.mimetype,
                "content": attachment.b64content,
            }
        )

    def set_extra_headers(self, headers):
        self.data.setdefault("additional_headers", []).extend(
            [{"key": key, "value": str(value)} for key, value in headers.items()]
        )

    def set_reply_to(self, emails):
        if emails:
            reply_to_string = ", ".join([str(email) for email in emails])
            self.data.setdefault("additional_headers", []).append(
                {"key": "Reply-To", "value": reply_to_string}
            )

    def set_tags(self, tags):
        if tags:
            self.data.setdefault("additional_headers", []).append(
                {"key": "X-Tags", "value": self.serialize_json(tags)}
            )

    def set_metadata(self, metadata):
        if metadata:
            self.data.setdefault("additional_headers", []).append(
                {"key": "X-Metadata", "value": self.serialize_json(metadata)}
            )

    def set_esp_extra(self, extra):
        self.data.update(extra)

    def serialize_data(self):
        return self.serialize_json(self.data)



import email.charset
import email.encoders
import email.policy

from requests.structures import CaseInsensitiveDict

from .. import __version__ as ANYMAIL_VERSION
from ..exceptions import AnymailAPIError, AnymailImproperlyInstalled
from ..message import AnymailRecipientStatus
from ..utils import UNSET, get_anymail_setting
from .base import AnymailBaseBackend, BasePayload

try:
    import boto3
    from botocore.client import Config
    from botocore.exceptions import BotoCoreError, ClientError, ConnectionError
except ImportError as err:
    raise AnymailImproperlyInstalled(
        missing_package="boto3", install_extra="amazon-ses"
    ) from err


# boto3 has several root exception classes; this is meant to cover all of them
BOTO_BASE_ERRORS = (BotoCoreError, ClientError, ConnectionError)


class EmailBackend(AnymailBaseBackend):
    """
    Amazon SES v2 Email Backend (using boto3)
    """

    esp_name = "Amazon SES"

    def __init__(self, **kwargs):
        """Init options from Django settings"""
        super().__init__(**kwargs)
        # AMAZON_SES_CLIENT_PARAMS is optional
        # (boto3 can find credentials several other ways)
        self.session_params, self.client_params = _get_anymail_boto3_params(
            esp_name=self.esp_name, kwargs=kwargs
        )
        self.configuration_set_name = get_anymail_setting(
            "configuration_set_name",
            esp_name=self.esp_name,
            kwargs=kwargs,
            allow_bare=False,
            default=None,
        )
        self.message_tag_name = get_anymail_setting(
            "message_tag_name",
            esp_name=self.esp_name,
            kwargs=kwargs,
            allow_bare=False,
            default=None,
        )

        self.client = None

    def open(self):
        if self.client:
            return False  # already exists
        try:
            self.client = boto3.session.Session(**self.session_params).client(
                "sesv2", **self.client_params
            )
        except Exception:
            if not self.fail_silently:
                raise
        else:
            return True  # created client

    def close(self):
        if self.client is None:
            return
        self.client.close()
        self.client = None

    def _send(self, message):
        if self.client:
            return super()._send(message)
        elif self.fail_silently:
            # (Probably missing boto3 credentials in open().)
            return False
        else:
            class_name = self.__class__.__name__
            raise RuntimeError(
                "boto3 Session has not been opened in {class_name}._send. "
                "(This is either an implementation error in {class_name}, "
                "or you are incorrectly calling _send directly.)".format(
                    class_name=class_name
                )
            )

    def build_message_payload(self, message, defaults):
        if getattr(message, "template_id", UNSET) is not UNSET:
            # For simplicity, use SESv2 SendBulkEmail for all templated messages
            # (even though SESv2 SendEmail has a template option).
            return AmazonSESV2SendBulkEmailPayload(message, defaults, self)
        else:
            return AmazonSESV2SendEmailPayload(message, defaults, self)

    def post_to_esp(self, payload, message):
        payload.finalize_payload()
        try:
            client_send_api = getattr(self.client, payload.api_name)
        except AttributeError:
            raise NotImplementedError(
                f"{self.client!r} does not have method {payload.api_name!r}."
            ) from None
        try:
            response = client_send_api(**payload.params)
        except BOTO_BASE_ERRORS as err:
            # ClientError has a response attr with parsed json error response
            # (other errors don't)
            raise AnymailAPIError(
                str(err),
                backend=self,
                email_message=message,
                payload=payload,
                response=getattr(err, "response", None),
            ) from err
        return response

    def parse_recipient_status(self, response, payload, message):
        return payload.parse_recipient_status(response)


class AmazonSESBasePayload(BasePayload):
    #: Name of the boto3 SES/SESv2 client method to call
    api_name = "SUBCLASS_MUST_OVERRIDE"

    def init_payload(self):
        self.params = {}
        if self.backend.configuration_set_name is not None:
            self.params["ConfigurationSetName"] = self.backend.configuration_set_name

    def finalize_payload(self):
        pass

    def parse_recipient_status(self, response):
        # response is the parsed (dict) JSON returned from the API call
        raise NotImplementedError()

    def set_esp_extra(self, extra):
        # e.g., ConfigurationSetName, FromEmailAddressIdentityArn,
        # FeedbackForwardingEmailAddress, ListManagementOptions
        self.params.update(extra)


class AmazonSESV2SendEmailPayload(AmazonSESBasePayload):
    api_name = "send_email"

    def init_payload(self):
        super().init_payload()
        self.all_recipients = []  # for parse_recipient_status
        self.mime_message = self.message.message()

    def finalize_payload(self):
        # (The boto3 SES client handles base64 encoding raw_message.)
        raw_message = self.generate_raw_message()
        self.params["Content"] = {"Raw": {"Data": raw_message}}

    def generate_raw_message(self):
        """
        Serialize self.mime_message as an RFC-5322/-2045 MIME message,
        encoded as 7bit-clean, us-ascii byte data.
        """
        # Amazon SES discourages `Content-Transfer-Encoding: 8bit`. And using
        # 8bit with SES open or click tracking results in mis-encoded characters.
        # To avoid this, convert any 8bit parts to 7bit quoted printable or base64.
        # (We own self.mime_message, so destructively modifying it should be OK.)
        for part in self.mime_message.walk():
            if part["Content-Transfer-Encoding"] == "8bit":
                del part["Content-Transfer-Encoding"]
                if part.get_content_maintype() == "text":
                    # (Avoid base64 for text parts, which can trigger spam filters)
                    email.encoders.encode_quopri(part)
                else:
                    email.encoders.encode_base64(part)

        # (All message and part headers should already be 7bit clean,
        # so there's no need to try to override email.policy here.)
        return self.mime_message.as_bytes()

    def parse_recipient_status(self, response):
        try:
            message_id = response["MessageId"]
        except (KeyError, TypeError) as err:
            raise AnymailAPIError(
                f"{err!s} parsing Amazon SES send result {response!r}",
                backend=self.backend,
                email_message=self.message,
                payload=self,
            ) from None

        recipient_status = AnymailRecipientStatus(
            message_id=message_id, status="queued"
        )
        return {
            recipient.addr_spec: recipient_status for recipient in self.all_recipients
        }

    # Standard EmailMessage attrs...
    # These all get rolled into the RFC-5322 raw mime directly via
    # EmailMessage.message()

    def _no_send_defaults(self, attr):
        # Anymail global send defaults don't work for standard attrs, because the
        # merged/computed value isn't forced back into the EmailMessage.
        if attr in self.defaults:
            self.unsupported_feature(
                f"Anymail send defaults for '{attr}' with Amazon SES"
            )

    def set_from_email(self, email):
        # If params["FromEmailAddress"] is not provided, SES will parse it from the raw
        # mime_message headers. (And setting it replaces any From header. Note that
        # v2 SendEmail doesn't have an equivalent to v1 SendRawEmail's Sender param.)
        self._no_send_defaults("from_email")

    def set_recipients(self, recipient_type, emails):
        # Although Amazon SES can parse the 'to' and 'cc' recipients from the raw
        # mime_message headers, providing them in the Destination param makes it
        # explicit (and is required for 'bcc' and for spoofed 'to').
        self.all_recipients += emails  # save for parse_recipient_status
        self._no_send_defaults(recipient_type)

        if emails:
            # params["Destination"] = {"ToAddresses": [...], "CcAddresses": etc.}
            # (Unlike most SendEmail params, these _don't_ replace the corresponding
            # raw mime_message headers.)
            assert recipient_type in ("to", "cc", "bcc")
            destination_key = f"{recipient_type.capitalize()}Addresses"
            self.params.setdefault("Destination", {})[destination_key] = [
                email.address for email in emails
            ]

    def set_subject(self, subject):
        # included in mime_message
        self._no_send_defaults("subject")

    def set_reply_to(self, emails):
        # included in mime_message
        # (and setting params["ReplyToAddresses"] replaces any Reply-To header)
        self._no_send_defaults("reply_to")

    def set_extra_headers(self, headers):
        # included in mime_message
        self._no_send_defaults("extra_headers")

    def set_text_body(self, body):
        # included in mime_message
        self._no_send_defaults("body")

    def set_html_body(self, body):
        # included in mime_message
        self._no_send_defaults("body")

    def set_alternatives(self, alternatives):
        # included in mime_message
        self._no_send_defaults("alternatives")

    def set_attachments(self, attachments):
        # included in mime_message
        self._no_send_defaults("attachments")

    # Anymail-specific payload construction

    def set_envelope_sender(self, email):
        # Amazon SES will generate a unique mailfrom, and then forward any delivery
        # problem reports that address receives to the address specified here:
        self.params["FeedbackForwardingEmailAddress"] = email.addr_spec

    def set_spoofed_to_header(self, header_to):
        # django.core.mail.EmailMessage.message() has already set
        #   self.mime_message["To"] = header_to
        # and performed any necessary header sanitization.
        #
        # The actual "to" is already in params["Destination"]["ToAddresses"].
        #
        # So, nothing to do here, except prevent the default
        # "unsupported feature" error.
        pass

    def set_metadata(self, metadata):
        # Amazon SES has two mechanisms for adding custom data to a message:
        # * Custom message headers are available to webhooks (SNS notifications),
        #   but not in CloudWatch metrics/dashboards or Kinesis Firehose streams.
        #   Custom headers can be sent only with SendRawEmail.
        # * "Message Tags" are available to CloudWatch and Firehose, and to SNS
        #   notifications for SES *events* but not SES *notifications*. (Got that?)
        #   Message Tags also allow *very* limited characters in both name and value.
        #   Message Tags can be sent with any SES send call.
        # (See "How do message tags work?" in
        # https://aws.amazon.com/blogs/ses/introducing-sending-metrics/
        # and https://forums.aws.amazon.com/thread.jspa?messageID=782922.)
        # To support reliable retrieval in webhooks, just use custom headers for
        # metadata.
        self.mime_message["X-Metadata"] = self.serialize_json(metadata)

    def set_merge_headers(self, merge_headers):
        self.unsupported_feature("merge_headers without template_id")

    def set_tags(self, tags):
        # See note about Amazon SES Message Tags and custom headers in set_metadata
        # above. To support reliable retrieval in webhooks, use custom headers for tags.
        # (There are no restrictions on number or content for custom header tags.)
        for tag in tags:
            # creates multiple X-Tag headers, one per tag:
            self.mime_message.add_header("X-Tag", tag)

        # Also *optionally* pass a single Message Tag if the AMAZON_SES_MESSAGE_TAG_NAME
        # Anymail setting is set (default no). The AWS API restricts tag content in this
        # case. (This is useful for dashboard segmentation; use esp_extra["Tags"] for
        # anything more complex.)
        if tags and self.backend.message_tag_name is not None:
            if len(tags) > 1:
                self.unsupported_feature(
                    "multiple tags with the AMAZON_SES_MESSAGE_TAG_NAME setting"
                )
            self.params.setdefault("EmailTags", []).append(
                {"Name": self.backend.message_tag_name, "Value": tags[0]}
            )

    def set_template_id(self, template_id):
        raise NotImplementedError(
            f"{self.__class__.__name__} should not have been used with template_id"
        )

    def set_merge_data(self, merge_data):
        self.unsupported_feature("merge_data without template_id")

    def set_merge_global_data(self, merge_global_data):
        self.unsupported_feature("global_merge_data without template_id")


class AmazonSESV2SendBulkEmailPayload(AmazonSESBasePayload):
    api_name = "send_bulk_email"

    def init_payload(self):
        super().init_payload()
        # late-bind in finalize_payload:
        self.recipients = {"to": [], "cc": [], "bcc": []}
        self.merge_data = {}
        self.headers = {}
        self.merge_headers = {}
        self.metadata = {}
        self.merge_metadata = {}
        self.tags = []

    def finalize_payload(self):
        # Build BulkEmailEntries from recipients and merge_data.

        # Any cc and bcc recipients should be included in every entry:
        cc_and_bcc_addresses = {}
        if self.recipients["cc"]:
            cc_and_bcc_addresses["CcAddresses"] = [
                cc.address for cc in self.recipients["cc"]
            ]
        if self.recipients["bcc"]:
            cc_and_bcc_addresses["BccAddresses"] = [
                bcc.address for bcc in self.recipients["bcc"]
            ]

        # Construct an entry with merge data for each "to" recipient:
        self.params["BulkEmailEntries"] = []
        for to in self.recipients["to"]:
            entry = {
                "Destination": dict(ToAddresses=[to.address], **cc_and_bcc_addresses),
                "ReplacementEmailContent": {
                    "ReplacementTemplate": {
                        "ReplacementTemplateData": self.serialize_json(
                            self.merge_data.get(to.addr_spec, {})
                        ),
                    }
                },
            }

            replacement_headers = []
            if self.headers or to.addr_spec in self.merge_headers:
                headers = CaseInsensitiveDict(self.headers)
                headers.update(self.merge_headers.get(to.addr_spec, {}))
                replacement_headers += [
                    {"Name": key, "Value": value} for key, value in headers.items()
                ]
            if self.metadata or to.addr_spec in self.merge_metadata:
                metadata = self.metadata.copy()
                metadata.update(self.merge_metadata.get(to.addr_spec, {}))
                if metadata:
                    replacement_headers.append(
                        {"Name": "X-Metadata", "Value": self.serialize_json(metadata)}
                    )
            if self.tags:
                replacement_headers += [
                    {"Name": "X-Tag", "Value": tag} for tag in self.tags
                ]
            if replacement_headers:
                entry["ReplacementHeaders"] = replacement_headers
            self.params["BulkEmailEntries"].append(entry)

    def parse_recipient_status(self, response):
        try:
            results = response["BulkEmailEntryResults"]
            ses_status_set = set(result["Status"] for result in results)
            anymail_statuses = [
                AnymailRecipientStatus(
                    message_id=result.get("MessageId", None),
                    status="queued" if result["Status"] == "SUCCESS" else "failed",
                )
                for result in results
            ]
        except (KeyError, TypeError) as err:
            raise AnymailAPIError(
                f"{err!s} parsing Amazon SES send result {response!r}",
                backend=self.backend,
                email_message=self.message,
                payload=self,
            ) from None

        # If all BulkEmailEntryResults[].Status are the same non-success status,
        # raise an APIError to expose the error message/reason (matching behavior
        # of non-template SendEmail call).
        if len(ses_status_set) == 1 and ses_status_set != {"SUCCESS"}:
            raise AnymailAPIError(
                # use Error text if available, else the Status enum, from first result
                results[0].get("Error", results[0]["Status"]),
                backend=self.backend,
                email_message=self.message,
                payload=self,
                response=response,
            )

        # Otherwise, return per-recipient status (just "queued" or "failed") for
        # all-success, mixed success/error, or all-error mixed-reason cases.
        # The BulkEmailEntryResults are in the same order as the Destination param
        # (which is in the same order as recipients["to"]).
        to_addrs = [to.addr_spec for to in self.recipients["to"]]
        if len(anymail_statuses) != len(to_addrs):
            raise AnymailAPIError(
                f"Sent to {len(to_addrs)} destinations,"
                f" but only {len(anymail_statuses)} statuses"
                f" in Amazon SES send result {response!r}",
                backend=self.backend,
                email_message=self.message,
                payload=self,
            )
        return dict(zip(to_addrs, anymail_statuses))

    def set_from_email(self, email):
        # this will RFC2047-encode display_name if needed:
        self.params["FromEmailAddress"] = email.address

    def set_recipients(self, recipient_type, emails):
        # late-bound in finalize_payload
        assert recipient_type in ("to", "cc", "bcc")
        self.recipients[recipient_type] = emails

    def set_subject(self, subject):
        # (subject can only come from template; you can use substitution vars in that)
        if subject:
            self.unsupported_feature("overriding template subject")

    def set_reply_to(self, emails):
        if emails:
            self.params["ReplyToAddresses"] = [email.address for email in emails]

    def set_extra_headers(self, headers):
        self.headers = headers

    def set_text_body(self, body):
        if body:
            self.unsupported_feature("overriding template body content")

    def set_html_body(self, body):
        if body:
            self.unsupported_feature("overriding template body content")

    def set_attachments(self, attachments):
        if attachments:
            self.unsupported_feature("attachments with template")

    # Anymail-specific payload construction

    def set_envelope_sender(self, email):
        # Amazon SES will generate a unique mailfrom, and then forward any delivery
        # problem reports that address receives to the address specified here:
        self.params["FeedbackForwardingEmailAddress"] = email.addr_spec

    def set_metadata(self, metadata):
        self.metadata = metadata

    def set_merge_metadata(self, merge_metadata):
        self.merge_metadata = merge_metadata

    def set_tags(self, tags):
        self.tags = tags

        # Also *optionally* pass a single Message Tag if the AMAZON_SES_MESSAGE_TAG_NAME
        # Anymail setting is set (default no). The AWS API restricts tag content in this
        # case. (This is useful for dashboard segmentation; use esp_extra["Tags"] for
        # anything more complex.)
        if tags and self.backend.message_tag_name is not None:
            if len(tags) > 1:
                self.unsupported_feature(
                    "multiple tags with the AMAZON_SES_MESSAGE_TAG_NAME setting"
                )
            self.params["DefaultEmailTags"] = [
                {"Name": self.backend.message_tag_name, "Value": tags[0]}
            ]

    def set_template_id(self, template_id):
        # DefaultContent.Template.TemplateName
        self.params.setdefault("DefaultContent", {}).setdefault("Template", {})[
            "TemplateName"
        ] = template_id

    def set_merge_data(self, merge_data):
        # late-bound in finalize_payload
        self.merge_data = merge_data

    def set_merge_headers(self, merge_headers):
        # late-bound in finalize_payload
        self.merge_headers = merge_headers

    def set_merge_global_data(self, merge_global_data):
        # DefaultContent.Template.TemplateData
        self.params.setdefault("DefaultContent", {}).setdefault("Template", {})[
            "TemplateData"
        ] = self.serialize_json(merge_global_data)


def _get_anymail_boto3_params(esp_name=EmailBackend.esp_name, kwargs=None):
    """Returns 2 dicts of params for boto3.session.Session() and .client()

    Incorporates ANYMAIL["AMAZON_SES_SESSION_PARAMS"] and
    ANYMAIL["AMAZON_SES_CLIENT_PARAMS"] settings.

    Converts config dict to botocore.client.Config if needed

    May remove keys from kwargs, but won't modify original settings
    """
    # (shared with ..webhooks.amazon_ses)
    session_params = get_anymail_setting(
        "session_params", esp_name=esp_name, kwargs=kwargs, default={}
    )
    client_params = get_anymail_setting(
        "client_params", esp_name=esp_name, kwargs=kwargs, default={}
    )

    # Add Anymail user-agent, and convert config dict to botocore.client.Config
    client_params = client_params.copy()  # don't modify source
    config = Config(
        user_agent_extra="django-anymail/{version}-{esp}".format(
            esp=esp_name.lower().replace(" ", "-"), version=ANYMAIL_VERSION
        )
    )
    if "config" in client_params:
        # convert config dict to botocore.client.Config if needed
        client_params_config = client_params["config"]
        if not isinstance(client_params_config, Config):
            client_params_config = Config(**client_params_config)
        config = config.merge(client_params_config)
    client_params["config"] = config

    return session_params, client_params



import json
from io import BytesIO
from textwrap import dedent
from unittest.mock import ANY

from django.test import ignore_warnings, tag

from anymail.exceptions import AnymailNotSupportedWarning
from anymail.inbound import AnymailInboundMessage
from anymail.signals import AnymailInboundEvent
from anymail.webhooks.sendgrid import SendGridInboundWebhookView

from .utils import (
    dedent_bytes,
    encode_multipart,
    make_fileobj,
    sample_email_content,
    sample_image_content,
)
from .webhook_cases import WebhookTestCase


@tag("sendgrid")
@ignore_warnings(category=AnymailNotSupportedWarning)
class SendgridInboundTestCase(WebhookTestCase):
    def test_not_supported_warning(self):
        with self.assertWarns(
            AnymailNotSupportedWarning,
            msg="django-anymail has dropped official support for SendGrid.",
        ):
            self.client.post("/anymail/sendgrid/inbound/", data={"email": ""})

    def test_inbound_basics(self):
        raw_event = {
            "headers": dedent(
                """\
                Received: from mail.example.org by mx987654321.sendgrid.net ...
                Received: by mail.example.org for <test@inbound.example.com> ...
                DKIM-Signature: v=1; a=rsa-sha256; c=relaxed/relaxed; d=example.org; ...
                MIME-Version: 1.0
                Received: by 10.10.1.71 with HTTP; Wed, 11 Oct 2017 18:31:04 -0700 (PDT)
                From: "Displayed From" <from+test@example.org>
                Date: Wed, 11 Oct 2017 18:31:04 -0700
                Message-ID: <CAEPk3R+4Zr@mail.example.org>
                Subject: Test subject
                To: "Test Inbound" <test@inbound.example.com>, other@example.com
                Cc: cc@example.com
                Content-Type: multipart/mixed; boundary="94eb2c115edcf35387055b61f849"
                """
            ),
            "from": "Displayed From <from+test@example.org>",
            "to": "Test Inbound <test@inbound.example.com>, other@example.com",
            "subject": "Test subject",
            "text": "Test body plain",
            "html": "<div>Test body html</div>",
            "attachments": "0",
            "charsets": '{"to":"UTF-8","html":"UTF-8",'
            '"subject":"UTF-8","from":"UTF-8","text":"UTF-8"}',
            "envelope": '{"to":["test@inbound.example.com"],'
            '"from":"envelope-from@example.org"}',
            "sender_ip": "10.10.1.71",
            # yep, SendGrid uses not-exactly-json for this field:
            "dkim": "{@example.org : pass}",
            "SPF": "pass",
            "spam_score": "1.7",
            "spam_report": "Spam detection software, running on the system"
            ' "mx987654321.sendgrid.net", '
            "has identified this incoming email as possible spam...",
        }
        response = self.client.post("/anymail/sendgrid/inbound/", data=raw_event)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.inbound_handler,
            sender=SendGridInboundWebhookView,
            event=ANY,
            esp_name="SendGrid",
        )
        # AnymailInboundEvent
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailInboundEvent)
        self.assertEqual(event.event_type, "inbound")
        self.assertIsNone(event.timestamp)
        self.assertIsNone(event.event_id)
        self.assertIsInstance(event.message, AnymailInboundMessage)
        # esp_event is a Django HttpRequest:
        self.assertEqual(event.esp_event.POST.dict(), raw_event)

        # AnymailInboundMessage - convenience properties
        message = event.message

        self.assertEqual(message.from_email.display_name, "Displayed From")
        self.assertEqual(message.from_email.addr_spec, "from+test@example.org")
        self.assertEqual(
            [str(e) for e in message.to],
            ["Test Inbound <test@inbound.example.com>", "other@example.com"],
        )
        self.assertEqual([str(e) for e in message.cc], ["cc@example.com"])
        self.assertEqual(message.subject, "Test subject")
        self.assertEqual(message.date.isoformat(" "), "2017-10-11 18:31:04-07:00")
        self.assertEqual(message.text, "Test body plain")
        self.assertEqual(message.html, "<div>Test body html</div>")

        self.assertEqual(message.envelope_sender, "envelope-from@example.org")
        self.assertEqual(message.envelope_recipient, "test@inbound.example.com")
        self.assertIsNone(message.stripped_text)
        self.assertIsNone(message.stripped_html)
        # SendGrid doesn't give a simple spam yes/no; check the score yourself:
        self.assertIsNone(message.spam_detected)
        self.assertEqual(message.spam_score, 1.7)

        # AnymailInboundMessage - other headers
        self.assertEqual(message["Message-ID"], "<CAEPk3R+4Zr@mail.example.org>")
        self.assertEqual(
            message.get_all("Received"),
            [
                "from mail.example.org by mx987654321.sendgrid.net ...",
                "by mail.example.org for <test@inbound.example.com> ...",
                "by 10.10.1.71 with HTTP; Wed, 11 Oct 2017 18:31:04 -0700 (PDT)",
            ],
        )

    def test_attachments(self):
        att1 = BytesIO("test attachment".encode("utf-8"))
        att1.name = "test.txt"
        image_content = sample_image_content()
        att2 = BytesIO(image_content)
        att2.name = "image.png"
        email_content = sample_email_content()
        att3 = BytesIO(email_content)
        att3.name = "\\share\\mail\\forwarded.msg"
        att3.content_type = 'message/rfc822; charset="us-ascii"'
        raw_event = {
            "headers": "",
            "attachments": "3",
            "attachment-info": json.dumps(
                {
                    "attachment3": {
                        "filename": "\\share\\mail\\forwarded.msg",
                        "charset": "US-ASCII",
                        "type": "message/rfc822",
                    },
                    "attachment2": {
                        "filename": "image.png",
                        "type": "image/png",
                        "content-id": "abc123",
                    },
                    "attachment1": {
                        "filename": "test.txt",
                        "charset": "UTF-8",
                        "type": "text/plain",
                    },
                }
            ),
            "content-ids": '{"abc123": "attachment2"}',
            "attachment1": att1,
            "attachment2": att2,  # inline
            "attachment3": att3,
        }

        response = self.client.post("/anymail/sendgrid/inbound/", data=raw_event)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.inbound_handler,
            sender=SendGridInboundWebhookView,
            event=ANY,
            esp_name="SendGrid",
        )
        event = kwargs["event"]
        message = event.message
        attachments = message.attachments  # AnymailInboundMessage convenience accessor
        self.assertEqual(len(attachments), 2)
        self.assertEqual(attachments[0].get_filename(), "test.txt")
        self.assertEqual(attachments[0].get_content_type(), "text/plain")
        self.assertEqual(attachments[0].get_content_text(), "test attachment")
        # Django strips path:
        self.assertEqual(attachments[1].get_filename(), "forwarded.msg")
        self.assertEqual(attachments[1].get_content_type(), "message/rfc822")
        self.assertEqualIgnoringHeaderFolding(
            attachments[1].get_content_bytes(), email_content
        )

        inlines = message.content_id_map
        self.assertEqual(len(inlines), 1)
        inline = inlines["abc123"]
        self.assertEqual(inline.get_filename(), "image.png")
        self.assertEqual(inline.get_content_type(), "image/png")
        self.assertEqual(inline.get_content_bytes(), image_content)

    def test_filtered_attachment_filenames(self):
        # Make sure the inbound webhook can deal with missing fields caused by
        # Django's multipart/form-data filename filtering. (The attachments are lost,
        # but shouldn't cause errors in the inbound webhook.)
        filenames = [
            "",
            "path\\",
            "path/" ".",
            "path\\.",
            "path/.",
            "..",
            "path\\..",
            "path/..",
        ]
        num_attachments = len(filenames)
        payload = {
            "attachment%d"
            % (i + 1): make_fileobj(
                "content", filename=filenames[i], content_type="text/pdf"
            )
            for i in range(num_attachments)
        }
        attachment_info = {
            key: {"filename": value.name, "type": "text/pdf"}
            for key, value in payload.items()
        }
        payload.update(
            {
                "headers": "",
                "attachments": str(num_attachments),
                "attachment-info": json.dumps(attachment_info),
            }
        )

        # Must do our own form-data encoding to properly test empty attachment
        # filenames. Must do our own multipart/form-data encoding for empty filenames:
        response = self.client.post(
            "/anymail/sendgrid/inbound/",
            data=encode_multipart("BoUnDaRy", payload),
            content_type="multipart/form-data; boundary=BoUnDaRy",
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.inbound_handler,
            sender=SendGridInboundWebhookView,
            event=ANY,
            esp_name="SendGrid",
        )

        # Different Django releases strip different filename patterns.
        # Just verify that at least some attachments got dropped (so the test is valid)
        # without causing an error in the inbound webhook:
        attachments = kwargs["event"].message.attachments
        self.assertLess(len(attachments), num_attachments)

    def test_inbound_mime(self):
        # SendGrid has an option to send the full, raw MIME message
        raw_event = {
            "email": dedent(
                """\
                From: A tester <test@example.org>
                Date: Thu, 12 Oct 2017 18:03:30 -0700
                Message-ID: <CAEPk3RKEx@mail.example.org>
                Subject: Raw MIME test
                To: test@inbound.example.com
                MIME-Version: 1.0
                Content-Type: multipart/alternative; boundary="94eb2c05e174adb140055b6339c5"

                --94eb2c05e174adb140055b6339c5
                Content-Type: text/plain; charset="UTF-8"
                Content-Transfer-Encoding: quoted-printable

                It's a body=E2=80=A6

                --94eb2c05e174adb140055b6339c5
                Content-Type: text/html; charset="UTF-8"
                Content-Transfer-Encoding: quoted-printable

                <div dir=3D"ltr">It's a body=E2=80=A6</div>

                --94eb2c05e174adb140055b6339c5--
                """  # NOQA: E501
            ),
            "from": "A tester <test@example.org>",
            "to": "test@inbound.example.com",
            "subject": "Raw MIME test",
            "charsets": '{"to":"UTF-8","subject":"UTF-8","from":"UTF-8"}',
            "envelope": '{"to":["test@inbound.example.com"],'
            '"from":"envelope-from@example.org"}',
            "sender_ip": "10.10.1.71",
            # yep, SendGrid uses not-exactly-json for this field:
            "dkim": "{@example.org : pass}",
            "SPF": "pass",
            "spam_score": "1.7",
            "spam_report": "Spam detection software, running on the system"
            ' "mx987654321.sendgrid.net", '
            "has identified this incoming email as possible spam...",
        }

        response = self.client.post("/anymail/sendgrid/inbound/", data=raw_event)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.inbound_handler,
            sender=SendGridInboundWebhookView,
            event=ANY,
            esp_name="SendGrid",
        )
        event = kwargs["event"]
        message = event.message
        self.assertEqual(message.envelope_sender, "envelope-from@example.org")
        self.assertEqual(message.envelope_recipient, "test@inbound.example.com")
        self.assertEqual(message.subject, "Raw MIME test")
        self.assertEqual(message.text, "It's a body\N{HORIZONTAL ELLIPSIS}\n")
        self.assertEqual(
            message.html,
            """<div dir="ltr">It's a body\N{HORIZONTAL ELLIPSIS}</div>\n""",
        )

    def test_inbound_charsets(self):
        # Captured (sanitized) from actual SendGrid inbound webhook payload 7/2020,
        # using a test message constructed with a variety of charsets:
        raw_post = dedent_bytes(
            b"""\
            --xYzZY
            Content-Disposition: form-data; name="headers"

            Date: Fri, 24 Jul 2020 16:43:46 UTC
            To: =?utf-8?q?R=C3=A9cipiendaire_pr=C3=A9cieux?= <inbound@sg.example.com>
            From: =?utf-8?q?Op=C3=A9rateur?= de test <sender@example.com>
            Subject: =?cp850?q?Como_usted_pidi=A2?=

            --xYzZY
            Content-Disposition: form-data; name="subject"

            Como usted pidi\xa2
            --xYzZY
            Content-Disposition: form-data; name="to"

            R\xc3\xa9cipiendaire pr\xc3\xa9cieux <inbound@sg.example.com>
            --xYzZY
            Content-Disposition: form-data; name="html"

            <p>\xbfEsto se ve como esperabas?</p>
            --xYzZY
            Content-Disposition: form-data; name="from"

            Op\xc3\xa9rateur de test <sender@example.com>
            --xYzZY
            Content-Disposition: form-data; name="text"

            Test the ESP\x92s inbound charset handling\x85
            --xYzZY
            Content-Disposition: form-data; name="charsets"

            {"to":"UTF-8","cc":"UTF-8","html":"iso-8859-1","subject":"cp850","from":"UTF-8","text":"windows-1252"}
            --xYzZY--
            """
        ).replace(b"\n", b"\r\n")

        response = self.client.post(
            "/anymail/sendgrid/inbound/",
            data=raw_post,
            content_type="multipart/form-data; boundary=xYzZY",
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.inbound_handler,
            sender=SendGridInboundWebhookView,
            event=ANY,
            esp_name="SendGrid",
        )
        event = kwargs["event"]
        message = event.message

        self.assertEqual(message.from_email.display_name, "Opérateur de test")
        self.assertEqual(message.from_email.addr_spec, "sender@example.com")
        self.assertEqual(len(message.to), 1)
        self.assertEqual(message.to[0].display_name, "Récipiendaire précieux")
        self.assertEqual(message.to[0].addr_spec, "inbound@sg.example.com")
        self.assertEqual(message.subject, "Como usted pidió")
        self.assertEqual(message.text, "Test the ESP’s inbound charset handling…")
        self.assertEqual(message.html, "<p>¿Esto se ve como esperabas?</p>")



import json
from datetime import datetime, timezone
from unittest.mock import ANY

from django.test import override_settings, tag

from anymail.signals import AnymailTrackingEvent
from anymail.webhooks.sparkpost import SparkPostTrackingWebhookView

from .webhook_cases import WebhookBasicAuthTestCase, WebhookTestCase


@tag("sparkpost")
class SparkPostWebhookSecurityTestCase(WebhookBasicAuthTestCase):
    def call_webhook(self):
        return self.client.post(
            "/anymail/sparkpost/tracking/",
            content_type="application/json",
            data=json.dumps([]),
        )

    # Actual tests are in WebhookBasicAuthTestCase


@tag("sparkpost")
class SparkPostDeliveryTestCase(WebhookTestCase):
    def test_ping_event(self):
        raw_events = [{"msys": {}}]
        response = self.client.post(
            "/anymail/sparkpost/tracking/",
            content_type="application/json",
            data=json.dumps(raw_events),
        )
        self.assertEqual(response.status_code, 200)
        self.assertFalse(self.tracking_handler.called)  # no real events

    def test_injection_event(self):
        # Full event from SparkPost sample events API.
        # (Later tests omit unused event fields.)
        raw_events = [
            {
                "msys": {
                    "message_event": {
                        "type": "injection",
                        "campaign_id": "Example Campaign Name",
                        "customer_id": "1",
                        "event_id": "92356927693813856",
                        "friendly_from": "sender@example.com",
                        "ip_pool": "Example-Ip-Pool",
                        "message_id": "000443ee14578172be22",
                        "msg_from": "sender@example.com",
                        "msg_size": "1337",
                        "rcpt_meta": {"customKey": "customValue"},
                        "rcpt_tags": ["male", "US"],
                        "rcpt_to": "recipient@example.com",
                        "raw_rcpt_to": "recipient@example.com",
                        "rcpt_type": "cc",
                        "routing_domain": "example.com",
                        "sending_ip": "127.0.0.1",
                        "sms_coding": "ASCII",
                        "sms_dst": "7876712656",
                        "sms_dst_npi": "E164",
                        "sms_dst_ton": "International",
                        "sms_segments": 5,
                        "sms_src": "1234",
                        "sms_src_npi": "E164",
                        "sms_src_ton": "Unknown",
                        "sms_text": "lol",
                        "subaccount_id": "101",
                        "subject": "Summer deals are here!",
                        "template_id": "templ-1234",
                        "template_version": "1",
                        "timestamp": "1454442600",
                        "transmission_id": "65832150921904138",
                    }
                }
            }
        ]
        response = self.client.post(
            "/anymail/sparkpost/tracking/",
            content_type="application/json",
            data=json.dumps(raw_events),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=SparkPostTrackingWebhookView,
            event=ANY,
            esp_name="SparkPost",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "queued")
        self.assertEqual(
            event.timestamp, datetime(2016, 2, 2, 19, 50, 00, tzinfo=timezone.utc)
        )
        self.assertEqual(event.esp_event, raw_events[0])
        # normalized "message_id" is actually transmission_id:
        self.assertEqual(event.message_id, "65832150921904138")
        self.assertEqual(event.event_id, "92356927693813856")
        self.assertEqual(event.recipient, "recipient@example.com")
        # campaign_id (rcpt_tags not available at send):
        self.assertEqual(event.tags, ["Example Campaign Name"])
        # includes transmissions.send metadata:
        self.assertEqual(event.metadata, {"customKey": "customValue"})

    def test_delivery_event(self):
        raw_events = [
            {
                "msys": {
                    "message_event": {
                        "type": "delivery",
                        "event_id": "92356927693813856",
                        "rcpt_to": "recipient@example.com",
                        "raw_rcpt_to": "Recipient@example.com",
                        "rcpt_meta": {},
                        "timestamp": "1454442600",
                        "transmission_id": "65832150921904138",
                    }
                }
            }
        ]
        response = self.client.post(
            "/anymail/sparkpost/tracking/",
            content_type="application/json",
            data=json.dumps(raw_events),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=SparkPostTrackingWebhookView,
            event=ANY,
            esp_name="SparkPost",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "delivered")
        self.assertEqual(event.recipient, "Recipient@example.com")
        self.assertEqual(event.tags, [])
        self.assertEqual(event.metadata, {})

    def test_bounce_event(self):
        raw_events = [
            {
                "msys": {
                    "message_event": {
                        "type": "bounce",
                        "bounce_class": "10",
                        "customer_id": "00000",
                        "error_code": "550",
                        "event_id": "84345317653491230",
                        "message_id": "0004e3724f57753a3561",
                        "raw_rcpt_to": "bounce@example.com",
                        "raw_reason": "550 5.1.1 <bounce@example.com>:"
                        " Recipient address rejected: User unknown",
                        "rcpt_to": "bounce@example.com",
                        "reason": "550 5.1.1 ...@... Recipient address rejected: ...",
                        "timestamp": "1464824548",
                        "transmission_id": "84345317650824116",
                    }
                },
                "cust": {"id": "00000"},  # Included in real (non-example) event data
            }
        ]
        response = self.client.post(
            "/anymail/sparkpost/tracking/",
            content_type="application/json",
            data=json.dumps(raw_events),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=SparkPostTrackingWebhookView,
            event=ANY,
            esp_name="SparkPost",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "bounced")
        self.assertEqual(event.esp_event, raw_events[0])
        self.assertEqual(event.message_id, "84345317650824116")  # transmission_id
        self.assertEqual(event.event_id, "84345317653491230")
        self.assertEqual(event.recipient, "bounce@example.com")
        self.assertEqual(event.reject_reason, "invalid")
        self.assertEqual(
            event.mta_response,
            "550 5.1.1 <bounce@example.com>: Recipient address rejected: User unknown",
        )

    def test_delay_event(self):
        raw_events = [
            {
                "msys": {
                    "message_event": {
                        "type": "delay",
                        "bounce_class": "21",
                        "error_code": "454",
                        "event_id": "84345317653675522",
                        "message_id": "0004e3724f57753a3861",
                        "num_retries": "1",
                        "queue_time": "1200161",
                        "raw_rcpt_to": "recipient@nomx.example.com",
                        "raw_reason": "454 4.4.4 [internal] no MX or A for domain",
                        "rcpt_to": "recipient@nomx.example.com",
                        "reason": "454 4.4.4 [internal] no MX or A for domain",
                        "timestamp": "1464825748",
                        "transmission_id": "84345317650824116",
                    }
                }
            }
        ]
        response = self.client.post(
            "/anymail/sparkpost/tracking/",
            content_type="application/json",
            data=json.dumps(raw_events),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=SparkPostTrackingWebhookView,
            event=ANY,
            esp_name="SparkPost",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "deferred")
        self.assertEqual(event.esp_event, raw_events[0])
        self.assertEqual(event.recipient, "recipient@nomx.example.com")
        self.assertEqual(
            event.mta_response, "454 4.4.4 [internal] no MX or A for domain"
        )

    def test_unsubscribe_event(self):
        raw_events = [
            {
                "msys": {
                    "unsubscribe_event": {
                        "type": "list_unsubscribe",
                        "event_id": "66331590532986193",
                        "message_id": "0004278150574660124d",
                        "raw_rcpt_to": "recipient@example.com",
                        "rcpt_to": "recipient@example.com",
                        "timestamp": "1464894280",
                        "transmission_id": "84345993965073285",
                    }
                }
            }
        ]
        response = self.client.post(
            "/anymail/sparkpost/tracking/",
            content_type="application/json",
            data=json.dumps(raw_events),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=SparkPostTrackingWebhookView,
            event=ANY,
            esp_name="SparkPost",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "unsubscribed")
        self.assertEqual(event.recipient, "recipient@example.com")

    def test_generation_rejection_event(self):
        # This is what you get if you try to send to a suppressed address
        raw_events = [
            {
                "msys": {
                    "gen_event": {
                        "type": "generation_rejection",
                        "error_code": "554",
                        "event_id": "102360394390563734",
                        "message_id": "0005c29950577c61695d",
                        "raw_rcpt_to": "suppressed@example.com",
                        "raw_reason": "554 5.7.1 recipient address suppressed"
                        " due to customer policy",
                        "rcpt_to": "suppressed@example.com",
                        "reason": "554 5.7.1 recipient address suppressed"
                        " due to customer policy",
                        "timestamp": "1464900034",
                        "transmission_id": "102360394387646691",
                    }
                }
            }
        ]
        response = self.client.post(
            "/anymail/sparkpost/tracking/",
            content_type="application/json",
            data=json.dumps(raw_events),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=SparkPostTrackingWebhookView,
            event=ANY,
            esp_name="SparkPost",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "rejected")
        self.assertEqual(event.recipient, "suppressed@example.com")
        self.assertEqual(
            event.mta_response,
            "554 5.7.1 recipient address suppressed due to customer policy",
        )

    def test_generation_failure_event(self):
        # This is what you get from a template rendering failure
        raw_events = [
            {
                "msys": {
                    "message_event": {
                        "type": "generation_failure",
                        "error_code": "554",
                        "event_id": "139013368081587254",
                        "raw_rcpt_to": "recipient@example.com",
                        "raw_reason": "554 5.3.3 [internal]"
                        " Error while rendering part html: ...",
                        "rcpt_subs": {"name": "Alice", "order_no": "12345"},
                        "rcpt_to": "recipient@example.com",
                        "reason": "554 5.3.3 [internal]"
                        " Error while rendering part html: ...",
                        "tdate": "2018-10-11T23:24:45.000Z",
                        "template_id": "test-template",
                        "template_version": "3",
                        "transmission_id": "139013368081177607",
                        "timestamp": "2018-10-11T23:24:45.000+00:00",
                    }
                }
            }
        ]
        response = self.client.post(
            "/anymail/sparkpost/tracking/",
            content_type="application/json",
            data=json.dumps(raw_events),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=SparkPostTrackingWebhookView,
            event=ANY,
            esp_name="SparkPost",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "failed")
        self.assertEqual(event.recipient, "recipient@example.com")
        self.assertEqual(
            event.mta_response,
            "554 5.3.3 [internal] Error while rendering part html: ...",
        )

    def test_bounce_challenge_response(self):
        # Test for changing initial event_type based on bounce_class
        raw_events = [
            {
                "msys": {
                    "message_event": {
                        "type": "bounce",
                        "bounce_class": "60",
                        "raw_rcpt_to": "vacationing@example.com",
                        "rcpt_to": "vacationing@example.com",
                    }
                }
            }
        ]
        response = self.client.post(
            "/anymail/sparkpost/tracking/",
            content_type="application/json",
            data=json.dumps(raw_events),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=SparkPostTrackingWebhookView,
            event=ANY,
            esp_name="SparkPost",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "autoresponded")
        self.assertEqual(event.reject_reason, "other")
        self.assertEqual(event.recipient, "vacationing@example.com")

    def test_open_event(self):
        raw_events = [
            {
                "msys": {
                    "track_event": {
                        "type": "open",
                        "raw_rcpt_to": "recipient@example.com",
                        "user_agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_3)"
                        " AppleWebKit/537.36",
                    }
                }
            }
        ]
        response = self.client.post(
            "/anymail/sparkpost/tracking/",
            content_type="application/json",
            data=json.dumps(raw_events),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=SparkPostTrackingWebhookView,
            event=ANY,
            esp_name="SparkPost",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "opened")
        self.assertEqual(
            event.user_agent,
            "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_3) AppleWebKit/537.36",
        )

    @override_settings(ANYMAIL_SPARKPOST_TRACK_INITIAL_OPEN_AS_OPENED=True)
    def test_initial_open_event_as_opened(self):
        # Mapping SparkPost "initial_open" to Anymail normalized "opened" is opt-in
        # via a setting, for backwards compatibility and to avoid reporting duplicate
        # "opened" events when all SparkPost event types are enabled.
        raw_events = [
            {
                "msys": {
                    "track_event": {
                        "type": "initial_open",
                        "raw_rcpt_to": "recipient@example.com",
                        "user_agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_3)"
                        " AppleWebKit/537.36",
                    }
                }
            }
        ]
        response = self.client.post(
            "/anymail/sparkpost/tracking/",
            content_type="application/json",
            data=json.dumps(raw_events),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=SparkPostTrackingWebhookView,
            event=ANY,
            esp_name="SparkPost",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "opened")
        self.assertEqual(
            event.user_agent,
            "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_3) AppleWebKit/537.36",
        )

    def test_initial_open_event_as_unknown(self):
        # By default, SparkPost "initial_open" is *not* mapped to Anymail "opened".
        raw_events = [
            {
                "msys": {
                    "track_event": {
                        "type": "initial_open",
                        "raw_rcpt_to": "recipient@example.com",
                        "user_agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_3)"
                        " AppleWebKit/537.36",
                    }
                }
            }
        ]
        response = self.client.post(
            "/anymail/sparkpost/tracking/",
            content_type="application/json",
            data=json.dumps(raw_events),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=SparkPostTrackingWebhookView,
            event=ANY,
            esp_name="SparkPost",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "unknown")
        # Here's how to get the raw SparkPost event type:
        self.assertEqual(
            event.esp_event["msys"].get("track_event", {}).get("type"), "initial_open"
        )
        # Note that other Anymail normalized event properties are still available:
        self.assertEqual(
            event.user_agent,
            "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_3) AppleWebKit/537.36",
        )

    def test_click_event(self):
        raw_events = [
            {
                "msys": {
                    "track_event": {
                        "type": "amp_click",
                        "raw_rcpt_to": "recipient@example.com",
                        "target_link_name": "Example Link Name",
                        "target_link_url": "http://example.com",
                        "user_agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_3)"
                        " AppleWebKit/537.36",
                    }
                }
            }
        ]
        response = self.client.post(
            "/anymail/sparkpost/tracking/",
            content_type="application/json",
            data=json.dumps(raw_events),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=SparkPostTrackingWebhookView,
            event=ANY,
            esp_name="SparkPost",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "clicked")
        self.assertEqual(event.recipient, "recipient@example.com")
        self.assertEqual(
            event.user_agent,
            "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_3) AppleWebKit/537.36",
        )
        self.assertEqual(event.click_url, "http://example.com")

    def test_amp_events(self):
        raw_events = [
            {
                "msys": {
                    "track_event": {
                        "type": "amp_open",
                    }
                }
            },
            {
                "msys": {
                    "track_event": {
                        "type": "amp_initial_open",
                    }
                }
            },
            {
                "msys": {
                    "track_event": {
                        "type": "amp_click",
                    }
                }
            },
        ]
        response = self.client.post(
            "/anymail/sparkpost/tracking/",
            content_type="application/json",
            data=json.dumps(raw_events),
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(self.tracking_handler.call_count, 3)
        events = [
            kwargs["event"] for (args, kwargs) in self.tracking_handler.call_args_list
        ]
        self.assertEqual(events[0].event_type, "opened")
        # amp_initial_open is mapped to "unknown" by default:
        self.assertEqual(events[1].event_type, "unknown")
        self.assertEqual(events[2].event_type, "clicked")



import json
from datetime import datetime
from email.encoders import encode_7or8bit
from email.mime.application import MIMEApplication
from unittest.mock import ANY, patch

from django.core import mail
from django.core.mail import BadHeaderError
from django.test import SimpleTestCase, override_settings, tag

from anymail import __version__ as ANYMAIL_VERSION
from anymail.exceptions import AnymailAPIError, AnymailUnsupportedFeature
from anymail.inbound import AnymailInboundMessage
from anymail.message import AnymailMessage, attach_inline_image_file

from .utils import (
    SAMPLE_IMAGE_FILENAME,
    AnymailTestMixin,
    sample_image_content,
    sample_image_path,
)


@tag("amazon_ses")
@override_settings(EMAIL_BACKEND="anymail.backends.amazon_ses.EmailBackend")
class AmazonSESBackendMockAPITestCase(AnymailTestMixin, SimpleTestCase):
    """TestCase that uses the Amazon SES EmailBackend with a mocked boto3 client"""

    def setUp(self):
        super().setUp()

        # Mock boto3.session.Session().client('sesv2').send_raw_email (and any other
        # client operations). (We could also use botocore.stub.Stubber, but mock works
        # well with our test structure.)
        self.patch_boto3_session = patch(
            "anymail.backends.amazon_ses.boto3.session.Session", autospec=True
        )
        self.mock_session = self.patch_boto3_session.start()  # boto3.session.Session
        self.addCleanup(self.patch_boto3_session.stop)
        #: boto3.session.Session().client
        self.mock_client = self.mock_session.return_value.client
        #: boto3.session.Session().client('sesv2', ...)
        self.mock_client_instance = self.mock_client.return_value
        self.set_mock_response()

        # Simple message useful for many tests
        self.message = mail.EmailMultiAlternatives(
            "Subject", "Text Body", "from@example.com", ["to@example.com"]
        )

    DEFAULT_SEND_RESPONSE = {
        "MessageId": "1111111111111111-bbbbbbbb-3333-7777-aaaa-eeeeeeeeeeee-000000",
        "ResponseMetadata": {
            "RequestId": "900dd7f3-0399-4a1b-9d9f-bed91f46924a",
            "HTTPStatusCode": 200,
            "HTTPHeaders": {
                "date": "Tue, 21 Feb 2023 22:59:46 GMT",
                "content-type": "application/json",
                "content-length": "76",
                "connection": "keep-alive",
                "x-amzn-requestid": "900dd7f3-0399-4a1b-9d9f-bed91f46924a",
            },
            "RetryAttempts": 0,
        },
    }

    def set_mock_response(self, response=None, operation_name="send_email"):
        mock_operation = getattr(self.mock_client_instance, operation_name)
        mock_operation.return_value = response or self.DEFAULT_SEND_RESPONSE
        return mock_operation.return_value

    def set_mock_failure(self, response, operation_name="send_email"):
        from botocore.exceptions import ClientError

        mock_operation = getattr(self.mock_client_instance, operation_name)
        mock_operation.side_effect = ClientError(
            response, operation_name=operation_name
        )

    def get_session_params(self):
        if self.mock_session.call_args is None:
            raise AssertionError("boto3 Session was not created")
        (args, kwargs) = self.mock_session.call_args
        if args:
            raise AssertionError(
                "boto3 Session created with unexpected positional args %r" % args
            )
        return kwargs

    def get_client_params(self, service="sesv2"):
        """Returns kwargs params passed to mock boto3 client constructor

        Fails test if boto3 client wasn't constructed with named service
        """
        if self.mock_client.call_args is None:
            raise AssertionError("boto3 client was not created")
        (args, kwargs) = self.mock_client.call_args
        if len(args) != 1:
            raise AssertionError(
                "boto3 client created with unexpected positional args %r" % args
            )
        if args[0] != service:
            raise AssertionError(
                "boto3 client created with service %r, not %r" % (args[0], service)
            )
        return kwargs

    def get_send_params(self, operation_name="send_email"):
        """Returns kwargs params passed to the mock send API.

        Fails test if API wasn't called.
        """
        self.mock_client.assert_called_with("sesv2", config=ANY)
        mock_operation = getattr(self.mock_client_instance, operation_name)
        if mock_operation.call_args is None:
            raise AssertionError("API was not called")
        (args, kwargs) = mock_operation.call_args
        return kwargs

    def get_sent_message(self):
        """Returns a parsed version of the send_email Content.Raw.Data param"""
        params = self.get_send_params(operation_name="send_email")
        raw_mime = params["Content"]["Raw"]["Data"]
        parsed = AnymailInboundMessage.parse_raw_mime_bytes(raw_mime)
        return parsed

    def assert_esp_not_called(self, msg=None, operation_name="send_email"):
        mock_operation = getattr(self.mock_client_instance, operation_name)
        if mock_operation.called:
            raise AssertionError(msg or "ESP API was called and shouldn't have been")


@tag("amazon_ses")
class AmazonSESBackendStandardEmailTests(AmazonSESBackendMockAPITestCase):
    """Test backend support for Django standard email features"""

    def test_send_mail(self):
        """Test basic API for simple send"""
        mail.send_mail(
            "Subject here",
            "Here is the message.",
            "from@example.com",
            ["to@example.com"],
            fail_silently=False,
        )
        params = self.get_send_params()
        # send_email takes a fully-formatted MIME message.
        # This is a simple (if inexact) way to check for expected headers and body:
        raw_mime = params["Content"]["Raw"]["Data"]
        self.assertIsInstance(raw_mime, bytes)  # SendEmail expects Data as bytes
        self.assertIn(b"\nFrom: from@example.com\n", raw_mime)
        self.assertIn(b"\nTo: to@example.com\n", raw_mime)
        self.assertIn(b"\nSubject: Subject here\n", raw_mime)
        self.assertIn(b"\n\nHere is the message", raw_mime)
        # Destination must include all recipients:
        self.assertEqual(params["Destination"], {"ToAddresses": ["to@example.com"]})

    # Since the SES backend generates the MIME message using Django's
    # EmailMessage.message().to_string(), there's not really a need
    # to exhaustively test all the various standard email features.
    # (EmailMessage.message() is well tested in the Django codebase.)
    # Instead, just spot-check a few things...

    def test_destinations(self):
        self.message.to = ["to1@example.com", '"Recipient, second" <to2@example.com>']
        self.message.cc = ["cc1@example.com", "Also cc <cc2@example.com>"]
        self.message.bcc = ["bcc1@example.com", "BCC 2 <bcc2@example.com>"]
        self.message.send()
        params = self.get_send_params()
        self.assertEqual(
            params["Destination"],
            {
                "ToAddresses": [
                    "to1@example.com",
                    '"Recipient, second" <to2@example.com>',
                ],
                "CcAddresses": ["cc1@example.com", "Also cc <cc2@example.com>"],
                "BccAddresses": ["bcc1@example.com", "BCC 2 <bcc2@example.com>"],
            },
        )
        # Bcc's shouldn't appear in the message itself:
        self.assertNotIn(b"bcc", params["Content"]["Raw"]["Data"])

    def test_non_ascii_headers(self):
        self.message.subject = "Thử tin nhắn"  # utf-8 in subject header
        self.message.to = ['"Người nhận" <to@example.com>']  # utf-8 in display name
        self.message.cc = ["cc@thư.example.com"]  # utf-8 in domain
        self.message.send()
        params = self.get_send_params()
        raw_mime = params["Content"]["Raw"]["Data"]
        # Non-ASCII headers must use MIME encoded-word syntax:
        self.assertIn(b"\nSubject: =?utf-8?b?VGjhu60gdGluIG5o4bqvbg==?=\n", raw_mime)
        # Non-ASCII display names as well:
        self.assertIn(
            b"\nTo: =?utf-8?b?TmfGsOG7nWkgbmjhuq1u?= <to@example.com>\n", raw_mime
        )
        # Non-ASCII address domains must use Punycode:
        self.assertIn(b"\nCc: cc@xn--th-e0a.example.com\n", raw_mime)
        # SES doesn't support non-ASCII in the username@ part
        # (RFC 6531 "SMTPUTF8" extension)

        # Destinations must include all recipients (addr-spec only, must use Punycode):
        self.assertEqual(
            params["Destination"],
            {
                "ToAddresses": ["=?utf-8?b?TmfGsOG7nWkgbmjhuq1u?= <to@example.com>"],
                "CcAddresses": ["cc@xn--th-e0a.example.com"],
            },
        )

    def test_attachments(self):
        # These are \u2022 bullets ("\N{BULLET}") below:
        text_content = "• Item one\n• Item two\n• Item three"
        self.message.attach(
            filename="Une pièce jointe.txt",  # utf-8 chars in filename
            content=text_content,
            mimetype="text/plain",
        )

        # Should guess mimetype if not provided...
        png_content = b"PNG\xb4 pretend this is the contents of a png file"
        self.message.attach(filename="test.png", content=png_content)

        # Should work with a MIMEBase object (also tests no filename)...
        pdf_content = b"PDF\xb4 pretend this is valid pdf params"
        mimeattachment = MIMEApplication(pdf_content, "pdf")  # application/pdf
        mimeattachment["Content-Disposition"] = "attachment"
        self.message.attach(mimeattachment)

        self.message.send()
        sent_message = self.get_sent_message()
        attachments = sent_message.attachments
        self.assertEqual(len(attachments), 3)

        self.assertEqual(attachments[0].get_content_type(), "text/plain")
        self.assertEqual(attachments[0].get_filename(), "Une pièce jointe.txt")
        self.assertEqual(attachments[0].get_param("charset"), "utf-8")
        self.assertEqual(attachments[0].get_content_text(), text_content)

        self.assertEqual(attachments[1].get_content_type(), "image/png")
        # not inline:
        self.assertEqual(attachments[1].get_content_disposition(), "attachment")
        self.assertEqual(attachments[1].get_filename(), "test.png")
        self.assertEqual(attachments[1].get_content_bytes(), png_content)

        self.assertEqual(attachments[2].get_content_type(), "application/pdf")
        self.assertIsNone(attachments[2].get_filename())  # no filename specified
        self.assertEqual(attachments[2].get_content_bytes(), pdf_content)

    def test_embedded_images(self):
        image_filename = SAMPLE_IMAGE_FILENAME
        image_path = sample_image_path(image_filename)
        image_data = sample_image_content(image_filename)

        cid = attach_inline_image_file(self.message, image_path, domain="example.com")
        html_content = (
            '<p>This has an <img src="cid:%s" alt="inline" /> image.</p>' % cid
        )
        self.message.attach_alternative(html_content, "text/html")

        self.message.send()
        sent_message = self.get_sent_message()

        self.assertEqual(sent_message.html, html_content)

        inlines = sent_message.content_id_map
        self.assertEqual(len(inlines), 1)
        self.assertEqual(inlines[cid].get_content_type(), "image/png")
        self.assertEqual(inlines[cid].get_filename(), image_filename)
        self.assertEqual(inlines[cid].get_content_bytes(), image_data)

        # Make sure neither the html nor the inline image is treated as an attachment:
        params = self.get_send_params()
        raw_mime = params["Content"]["Raw"]["Data"]
        self.assertNotIn(b"\nContent-Disposition: attachment", raw_mime)

    def test_multiple_html_alternatives(self):
        # Multiple alternatives *are* allowed
        self.message.attach_alternative("<p>First html is OK</p>", "text/html")
        self.message.attach_alternative("<p>And so is second</p>", "text/html")
        self.message.send()
        params = self.get_send_params()
        raw_mime = params["Content"]["Raw"]["Data"]
        # just check the alternative smade it into the message
        # (assume that Django knows how to format them properly)
        self.assertIn(b"\n\n<p>First html is OK</p>\n", raw_mime)
        self.assertIn(b"\n\n<p>And so is second</p>\n", raw_mime)

    def test_alternative(self):
        # Non-HTML alternatives (including AMP) *are* allowed
        self.message.attach_alternative("<p>AMP HTML</p>", "text/x-amp-html")
        self.message.send()
        params = self.get_send_params()
        raw_mime = params["Content"]["Raw"]["Data"]
        # just check the alternative made it into the message
        # (assume that Python email knows how to format it properly)
        self.assertIn(b"\nContent-Type: text/x-amp-html", raw_mime)

    def test_multiple_from(self):
        # Amazon allows multiple addresses in the From header,
        # but must specify a single one for the FromEmailAddress
        self.message.from_email = "First <from1@example.com>, from2@example.com"
        with self.assertRaisesMessage(
            AnymailUnsupportedFeature, "multiple from emails"
        ):
            self.message.send()

    def test_commas_in_subject(self):
        """
        There used to be a Python email header bug that added unwanted spaces
        after commas in long subjects
        """
        self.message.subject = (
            "100,000,000 isn't a number you'd really want"
            " to break up in this email subject, right?"
        )
        self.message.send()
        sent_message = self.get_sent_message()
        self.assertEqual(sent_message["Subject"], self.message.subject)

    def test_broken_address_header(self):
        # https://github.com/anymail/django-anymail/issues/369
        self.message.to = ['"Người nhận a very very long, name" <to@example.com>']
        self.message.cc = [
            '"A véry long name with non-ASCII char and, comma" <cc@example.com>'
        ]
        self.message.send()
        sent_message = self.get_sent_message()
        self.assertEqual(sent_message["To"], self.message.to[0])
        self.assertEqual(sent_message["Cc"], self.message.cc[0])

    def test_no_cte_8bit(self):
        """Anymail works around an Amazon SES bug that can corrupt non-ASCII bodies."""
        # (see detailed comments in the backend code)

        # The generated MIMEText for each of these ends up using CTE 8bit by default:
        self.message.body = "Это text body"
        self.message.attach_alternative("<p>Это html body</p>", "text/html")
        self.message.attach("sample.csv", "Это attachment", "text/csv")

        # Also force a CTE 8bit attachment (which normally defaults to CTE base64):
        att = MIMEApplication("Это data".encode("utf8"), "data", encode_7or8bit)
        self.assertEqual(att["Content-Transfer-Encoding"], "8bit")
        self.message.attach(att)

        self.message.send()
        raw_mime = self.get_send_params()["Content"]["Raw"]["Data"]
        self.assertTrue(raw_mime.isascii())  # 7-bit clean

        # Make sure none of the resulting parts use `Content-Transfer-Encoding: 8bit`.
        # (Technically, either quoted-printable or base64 would be OK, but base64 text
        # parts have a reputation for triggering spam filters, so just require
        # quoted-printable for them.)
        sent_message = self.get_sent_message()
        part_encodings = [
            (part.get_content_type(), part["Content-Transfer-Encoding"])
            for part in sent_message.walk()
        ]
        self.assertEqual(
            part_encodings,
            [
                ("multipart/mixed", None),
                ("multipart/alternative", None),
                ("text/plain", "quoted-printable"),
                ("text/html", "quoted-printable"),
                ("text/csv", "quoted-printable"),
                ("application/data", "base64"),
            ],
        )

    def test_no_cte_8bit_root(self):
        # Same test as above, but with a non-multipart message using 8bit at root
        self.message.body = "Это text body"
        self.message.send()

        raw_mime = self.get_send_params()["Content"]["Raw"]["Data"]
        self.assertTrue(raw_mime.isascii())  # 7-bit clean

        sent_message = self.get_sent_message()
        part_encodings = [
            (part.get_content_type(), part["Content-Transfer-Encoding"])
            for part in sent_message.walk()
        ]
        self.assertEqual(part_encodings, [("text/plain", "quoted-printable")])

    def test_api_failure(self):
        error_response = {
            "Error": {
                "Code": "MessageRejected",
                "Message": "Email address is not verified. The following identities"
                " failed the check in region US-EAST-1: to@example.com",
            },
            "ResponseMetadata": {
                "RequestId": "c44b0ae2-e086-45ca-8820-b76a9b9f430a",
                "HTTPStatusCode": 403,
                "HTTPHeaders": {
                    "date": "Tue, 21 Feb 2023 23:49:31 GMT",
                    "content-type": "application/json",
                    "content-length": "196",
                    "connection": "keep-alive",
                    "x-amzn-requestid": "c44b0ae2-e086-45ca-8820-b76a9b9f430a",
                    "x-amzn-errortype": "MessageRejected",
                },
                "RetryAttempts": 0,
            },
        }

        self.set_mock_failure(error_response)
        with self.assertRaises(AnymailAPIError) as cm:
            self.message.send()
        err = cm.exception
        # AWS error is included in Anymail message:
        self.assertIn(
            "Email address is not verified. The following identities failed "
            "the check in region US-EAST-1: to@example.com",
            str(err),
        )
        # Raw AWS response is available on the exception:
        self.assertEqual(err.response, error_response)

    def test_api_failure_fail_silently(self):
        # Make sure fail_silently is respected
        self.set_mock_failure(
            {
                "Error": {
                    "Type": "Sender",
                    "Code": "InvalidParameterValue",
                    "Message": "That is not allowed",
                }
            }
        )
        sent = self.message.send(fail_silently=True)
        self.assertEqual(sent, 0)

    def test_session_failure_fail_silently(self):
        # Make sure fail_silently is respected if boto3.Session creation fails
        # (e.g., due to invalid or missing credentials)
        from botocore.exceptions import NoCredentialsError

        self.mock_session.side_effect = NoCredentialsError()

        sent = self.message.send(fail_silently=True)
        self.assertEqual(sent, 0)

    def test_prevents_header_injection(self):
        # Since we build the raw MIME message, we're responsible for preventing header
        # injection. django.core.mail.EmailMessage.message() implements most of that
        # (for the SMTP backend); spot check some likely cases just to be sure...
        with self.assertRaises(BadHeaderError):
            mail.send_mail(
                "Subject\r\ninjected", "Body", "from@example.com", ["to@example.com"]
            )
        with self.assertRaises(BadHeaderError):
            mail.send_mail(
                "Subject",
                "Body",
                '"Display-Name\nInjected" <from@example.com>',
                ["to@example.com"],
            )
        with self.assertRaises(BadHeaderError):
            mail.send_mail(
                "Subject",
                "Body",
                "from@example.com",
                ['"Display-Name\rInjected" <to@example.com>'],
            )
        with self.assertRaises(BadHeaderError):
            mail.EmailMessage(
                "Subject",
                "Body",
                "from@example.com",
                ["to@example.com"],
                headers={"X-Header": "custom header value\r\ninjected"},
            ).send()


@tag("amazon_ses")
class AmazonSESBackendAnymailFeatureTests(AmazonSESBackendMockAPITestCase):
    """Test backend support for Anymail added features"""

    def test_envelope_sender(self):
        self.message.envelope_sender = "bounce-handler@bounces.example.com"
        self.message.send()
        params = self.get_send_params()
        self.assertEqual(
            params["FeedbackForwardingEmailAddress"],
            "bounce-handler@bounces.example.com",
        )

    def test_spoofed_to(self):
        # Amazon SES is one of the few ESPs that actually permits the To header
        # to differ from the envelope recipient...
        self.message.to = ["Envelope <envelope-to@example.com>"]
        self.message.extra_headers["To"] = "Spoofed <spoofed-to@elsewhere.example.org>"
        self.message.send()
        params = self.get_send_params()
        raw_mime = params["Content"]["Raw"]["Data"]
        self.assertEqual(
            params["Destination"],
            {"ToAddresses": ["Envelope <envelope-to@example.com>"]},
        )
        self.assertIn(b"\nTo: Spoofed <spoofed-to@elsewhere.example.org>\n", raw_mime)
        self.assertNotIn(b"envelope-to@example.com", raw_mime)

    def test_metadata(self):
        self.message.metadata = {
            "User ID": 12345,
            # that \n is a header-injection test:
            "items": "Correct horse,Battery,\nStaple",
            "Cart-Total": "22.70",
        }
        self.message.send()

        # Metadata is passed as JSON in a message header field:
        sent_message = self.get_sent_message()
        self.assertJSONEqual(
            sent_message["X-Metadata"],
            '{"User ID": 12345,'
            ' "items": "Correct horse,Battery,\\nStaple",'
            ' "Cart-Total": "22.70"}',
        )

    def test_send_at(self):
        # Amazon SES does not support delayed sending
        self.message.send_at = datetime(2016, 3, 4, 5, 6, 7)
        with self.assertRaisesMessage(AnymailUnsupportedFeature, "send_at"):
            self.message.send()

    def test_tags(self):
        self.message.tags = ["Transactional", "Cohort 12/2017"]
        self.message.send()

        # Tags are added as multiple X-Tag message headers:
        sent_message = self.get_sent_message()
        self.assertCountEqual(
            sent_message.get_all("X-Tag"), ["Transactional", "Cohort 12/2017"]
        )

        # Tags are *not* by default used as Amazon SES "Message Tags":
        params = self.get_send_params()
        self.assertNotIn("Tags", params)

    @override_settings(ANYMAIL_AMAZON_SES_MESSAGE_TAG_NAME="Campaign")
    def test_amazon_message_tags(self):
        """
        The Anymail AMAZON_SES_MESSAGE_TAG_NAME setting enables a single Message Tag
        """
        self.message.tags = ["Welcome"]
        self.message.send()
        params = self.get_send_params()
        self.assertEqual(
            params["EmailTags"], [{"Name": "Campaign", "Value": "Welcome"}]
        )

        # Multiple Anymail tags are not supported when using this feature
        self.message.tags = ["Welcome", "Variation_A"]
        with self.assertRaisesMessage(
            AnymailUnsupportedFeature,
            "multiple tags with the AMAZON_SES_MESSAGE_TAG_NAME setting",
        ):
            self.message.send()

    def test_tracking(self):
        # Amazon SES doesn't support overriding click/open-tracking settings
        # on individual messages through any standard API params.
        # (You _can_ use a ConfigurationSet to control this; see esp_extra below.)
        self.message.track_clicks = True
        with self.assertRaisesMessage(AnymailUnsupportedFeature, "track_clicks"):
            self.message.send()
        delattr(self.message, "track_clicks")

        self.message.track_opens = True
        with self.assertRaisesMessage(AnymailUnsupportedFeature, "track_opens"):
            self.message.send()

    def test_merge_data(self):
        # Amazon SES only supports merging when using templates (see below)
        self.message.merge_data = {}
        with self.assertRaisesMessage(
            AnymailUnsupportedFeature, "merge_data without template_id"
        ):
            self.message.send()
        delattr(self.message, "merge_data")

        self.message.merge_global_data = {"group": "Users", "site": "ExampleCo"}
        with self.assertRaisesMessage(
            AnymailUnsupportedFeature, "global_merge_data without template_id"
        ):
            self.message.send()

    def test_merge_headers(self):
        # Amazon SES only supports merging when using templates (see below)
        self.message.merge_headers = {}
        with self.assertRaisesMessage(
            AnymailUnsupportedFeature, "merge_headers without template_id"
        ):
            self.message.send()

    def test_template(self):
        """With template_id, Anymail switches to SESv2 SendBulkEmail"""
        # SendBulkEmail uses a completely different API call and payload
        # structure, so this re-tests a bunch of Anymail features that were handled
        # differently above. (See test_amazon_ses_integration for a more realistic
        # template example.)
        raw_response = {
            "BulkEmailEntryResults": [
                {
                    "Status": "SUCCESS",
                    "MessageId": "1111111111111111-bbbbbbbb-3333-7777",
                },
                {
                    "Status": "ACCOUNT_DAILY_QUOTA_EXCEEDED",
                    "Error": "Daily message quota exceeded",
                },
            ],
            "ResponseMetadata": self.DEFAULT_SEND_RESPONSE["ResponseMetadata"],
        }
        self.set_mock_response(raw_response, operation_name="send_bulk_email")
        message = AnymailMessage(
            template_id="welcome_template",
            from_email='"Example, Inc." <from@example.com>',
            to=["alice@example.com", "罗伯特 <bob@example.com>"],
            cc=["cc@example.com"],
            reply_to=["reply1@example.com", "Reply 2 <reply2@example.com>"],
            headers={
                "List-Unsubscribe-Post": "List-Unsubscribe=One-Click",
                "List-Unsubscribe": "<mailto:unsubscribe@example.com>",
            },
            merge_headers={
                "alice@example.com": {
                    "List-Unsubscribe": "<https://example.com/a/>",
                },
                "bob@example.com": {
                    "List-Unsubscribe": "<https://example.com/b/>",
                },
            },
            merge_data={
                "alice@example.com": {"name": "Alice", "group": "Developers"},
                "bob@example.com": {"name": "Bob"},  # and leave group undefined
                "nobody@example.com": {"name": "Not a recipient for this message"},
            },
            merge_global_data={"group": "Users", "site": "ExampleCo"},
            tags=["Welcome Variant A", "Cohort 12/2017"],
            metadata={"meta1": "test"},
            merge_metadata={
                "alice@example.com": {"meta2": "meta-alice"},
            },
            envelope_sender="bounce@example.com",
            esp_extra={
                "FromEmailAddressIdentityArn": (
                    "arn:aws:ses:us-east-1:123456789012:identity/example.com"
                )
            },
        )
        message.send()

        # templates use a different API call...
        self.assert_esp_not_called(operation_name="send_email")
        params = self.get_send_params(operation_name="send_bulk_email")
        self.assertEqual(
            params["DefaultContent"]["Template"]["TemplateName"], "welcome_template"
        )
        self.assertEqual(
            params["FromEmailAddress"], '"Example, Inc." <from@example.com>'
        )
        bulk_entries = params["BulkEmailEntries"]
        self.assertEqual(len(bulk_entries), 2)
        self.assertEqual(
            bulk_entries[0]["Destination"],
            {"ToAddresses": ["alice@example.com"], "CcAddresses": ["cc@example.com"]},
        )
        self.assertEqual(
            json.loads(
                bulk_entries[0]["ReplacementEmailContent"]["ReplacementTemplate"][
                    "ReplacementTemplateData"
                ]
            ),
            {"name": "Alice", "group": "Developers"},
        )
        self.assertEqual(
            bulk_entries[1]["Destination"],
            {
                # SES requires RFC2047:
                "ToAddresses": ["=?utf-8?b?572X5Lyv54m5?= <bob@example.com>"],
                "CcAddresses": ["cc@example.com"],
            },
        )
        self.assertEqual(
            json.loads(
                bulk_entries[1]["ReplacementEmailContent"]["ReplacementTemplate"][
                    "ReplacementTemplateData"
                ]
            ),
            {"name": "Bob"},
        )

        self.assertCountEqual(
            bulk_entries[0]["ReplacementHeaders"],
            [
                # From extra_headers and merge_headers:
                {
                    "Name": "List-Unsubscribe-Post",
                    "Value": "List-Unsubscribe=One-Click",
                },
                {"Name": "List-Unsubscribe", "Value": "<https://example.com/a/>"},
                # From metadata and merge_metadata:
                {
                    "Name": "X-Metadata",
                    "Value": '{"meta1": "test", "meta2": "meta-alice"}',
                },
                # From tags:
                {"Name": "X-Tag", "Value": "Welcome Variant A"},
                {"Name": "X-Tag", "Value": "Cohort 12/2017"},
            ],
        )
        self.assertCountEqual(
            bulk_entries[1]["ReplacementHeaders"],
            [
                # From extra_headers and merge_headers:
                {
                    "Name": "List-Unsubscribe-Post",
                    "Value": "List-Unsubscribe=One-Click",
                },
                {"Name": "List-Unsubscribe", "Value": "<https://example.com/b/>"},
                # From metadata (no merge_metadata for bob@):
                {"Name": "X-Metadata", "Value": '{"meta1": "test"}'},
                # From tags:
                {"Name": "X-Tag", "Value": "Welcome Variant A"},
                {"Name": "X-Tag", "Value": "Cohort 12/2017"},
            ],
        )
        self.assertEqual(
            json.loads(params["DefaultContent"]["Template"]["TemplateData"]),
            {"group": "Users", "site": "ExampleCo"},
        )
        self.assertEqual(
            params["ReplyToAddresses"],
            ["reply1@example.com", "Reply 2 <reply2@example.com>"],
        )
        self.assertEqual(params["FeedbackForwardingEmailAddress"], "bounce@example.com")
        # esp_extra:
        self.assertEqual(
            params["FromEmailAddressIdentityArn"],
            "arn:aws:ses:us-east-1:123456789012:identity/example.com",
        )

        self.assertEqual(message.anymail_status.status, {"queued", "failed"})
        self.assertEqual(
            # different for each recipient
            message.anymail_status.message_id,
            {"1111111111111111-bbbbbbbb-3333-7777", None},
        )
        self.assertEqual(
            message.anymail_status.recipients["alice@example.com"].status, "queued"
        )
        self.assertEqual(
            message.anymail_status.recipients["bob@example.com"].status, "failed"
        )
        self.assertEqual(
            message.anymail_status.recipients["alice@example.com"].message_id,
            "1111111111111111-bbbbbbbb-3333-7777",
        )
        self.assertIsNone(
            message.anymail_status.recipients["bob@example.com"].message_id
        )
        self.assertEqual(message.anymail_status.esp_response, raw_response)

    def test_template_omits_unused_replacement_headers(self):
        """If headers are not needed, the ReplacementHeaders param should be omitted"""
        # bob@example.com requires ReplacementHeaders; alice@example.com doesn't
        raw_response = {
            "BulkEmailEntryResults": [
                {
                    "Status": "SUCCESS",
                    "MessageId": "1111111111111111-bbbbbbbb-3333-7777",
                },
                {
                    "Status": "SUCCESS",
                    "MessageId": "1111111111111111-bbbbbbbb-4444-8888",
                },
            ],
            "ResponseMetadata": self.DEFAULT_SEND_RESPONSE["ResponseMetadata"],
        }
        self.set_mock_response(raw_response, operation_name="send_bulk_email")
        message = AnymailMessage(
            template_id="welcome_template",
            from_email='"Example, Inc." <from@example.com>',
            to=["alice@example.com", "罗伯特 <bob@example.com>"],
            reply_to=["reply1@example.com", "Reply 2 <reply2@example.com>"],
            merge_headers={
                "alice@example.com": {},
                "bob@example.com": {"X-Test": "test"},
            },
            merge_global_data={"group": "Users", "site": "ExampleCo"},
        )
        message.send()

        params = self.get_send_params(operation_name="send_bulk_email")
        self.assertNotIn("ReplacementHeaders", params["BulkEmailEntries"][0])
        self.assertIn("ReplacementHeaders", params["BulkEmailEntries"][1])

    @override_settings(
        # This will pass DefaultEmailTags: Name "Campaign"
        ANYMAIL_AMAZON_SES_MESSAGE_TAG_NAME="Campaign"
    )
    def test_template_default_email_tag(self):
        raw_response = {
            "BulkEmailEntryResults": [
                {
                    "Status": "SUCCESS",
                    "MessageId": "1111111111111111-bbbbbbbb-3333-7777",
                },
            ],
            "ResponseMetadata": self.DEFAULT_SEND_RESPONSE["ResponseMetadata"],
        }
        self.set_mock_response(raw_response, operation_name="send_bulk_email")
        message = AnymailMessage(
            template_id="welcome_template",
            from_email='"Example, Inc." <from@example.com>',
            to=["alice@example.com"],
            tags=["WelcomeVariantA"],
        )
        message.send()

        params = self.get_send_params(operation_name="send_bulk_email")
        self.assertEqual(
            params["DefaultEmailTags"],
            [{"Name": "Campaign", "Value": "WelcomeVariantA"}],
        )

    def test_template_failure(self):
        """Failures to all recipients raise a similar error to non-template sends"""
        raw_response = {
            "BulkEmailEntryResults": [
                {
                    "Status": "TEMPLATE_DOES_NOT_EXIST",
                    "Error": "No template named 'oops'",
                },
                {
                    "Status": "TEMPLATE_DOES_NOT_EXIST",
                    "Error": "No template named 'oops'",
                },
            ],
            "ResponseMetadata": self.DEFAULT_SEND_RESPONSE["ResponseMetadata"],
        }
        self.set_mock_response(raw_response, operation_name="send_bulk_email")
        message = AnymailMessage(
            template_id="oops",
            from_email="from@example.com",
            to=["alice@example.com", "bob@example.com"],
        )
        with self.assertRaisesMessage(AnymailAPIError, "No template named 'oops'"):
            message.send()

    def test_template_unsupported(self):
        """Some options are not compatible with SendBulkTemplatedEmail"""
        message = AnymailMessage(template_id="welcome_template", to=["to@example.com"])

        message.subject = "nope, can't change template subject"
        with self.assertRaisesMessage(
            AnymailUnsupportedFeature, "overriding template subject"
        ):
            message.send()
        message.subject = None

        message.body = "nope, can't change text body"
        with self.assertRaisesMessage(
            AnymailUnsupportedFeature, "overriding template body content"
        ):
            message.send()
        message.content_subtype = "html"
        with self.assertRaisesMessage(
            AnymailUnsupportedFeature, "overriding template body content"
        ):
            message.send()
        message.body = None

        message.attach("attachment.txt", "this is an attachment", "text/plain")
        with self.assertRaisesMessage(
            AnymailUnsupportedFeature, "attachments with template"
        ):
            message.send()
        message.attachments = []

    def test_send_anymail_message_without_template(self):
        # Make sure SendEmail is used for non-template_id messages
        message = AnymailMessage(
            from_email="from@example.com", to=["to@example.com"], subject="subject"
        )
        message.send()
        self.assert_esp_not_called(operation_name="send_bulk_email")
        # fails if send_email not called:
        self.get_send_params(operation_name="send_email")

    def test_default_omits_options(self):
        """Make sure by default we don't send any ESP-specific options.

        Options not specified by the caller should be omitted entirely from
        the API call (*not* sent as False or empty). This ensures
        that your ESP account settings apply by default.
        """
        self.message.send()
        params = self.get_send_params()
        self.assertNotIn("BulkEmailEntries", params)
        self.assertNotIn("ConfigurationSetName", params)
        self.assertNotIn("DefaultContent", params)
        self.assertNotIn("DefaultContent", params)
        self.assertNotIn("DefaultEmailTags", params)
        self.assertNotIn("EmailTags", params)
        self.assertNotIn("FeedbackForwardingEmailAddress", params)
        self.assertNotIn("FeedbackForwardingEmailAddressIdentityArn", params)
        self.assertNotIn("FromEmailAddressIdentityArn", params)
        self.assertNotIn("ListManagementOptions", params)
        self.assertNotIn("ReplyToAddresses", params)

        sent_message = self.get_sent_message()
        # custom headers not added if not needed:
        self.assertNotIn("X-Metadata", sent_message)
        self.assertNotIn("X-Tag", sent_message)

    def test_esp_extra(self):
        # Values in esp_extra are merged into the Amazon SES SendRawEmail parameters
        self.message.esp_extra = {
            # E.g., if you've set up a configuration set
            # that disables open/click tracking:
            "ConfigurationSetName": "NoTrackingConfigurationSet",
        }
        self.message.send()
        params = self.get_send_params()
        self.assertEqual(params["ConfigurationSetName"], "NoTrackingConfigurationSet")

    def test_send_attaches_anymail_status(self):
        """The anymail_status should be attached to the message when it is sent"""
        msg = mail.EmailMessage(
            "Subject",
            "Message",
            "from@example.com",
            ["to1@example.com"],
        )
        sent = msg.send()
        self.assertEqual(sent, 1)
        self.assertEqual(msg.anymail_status.status, {"queued"})
        self.assertEqual(
            msg.anymail_status.message_id,
            "1111111111111111-bbbbbbbb-3333-7777-aaaa-eeeeeeeeeeee-000000",
        )
        self.assertEqual(
            msg.anymail_status.recipients["to1@example.com"].status, "queued"
        )
        self.assertEqual(
            msg.anymail_status.recipients["to1@example.com"].message_id,
            "1111111111111111-bbbbbbbb-3333-7777-aaaa-eeeeeeeeeeee-000000",
        )
        self.assertEqual(msg.anymail_status.esp_response, self.DEFAULT_SEND_RESPONSE)

    # Amazon SES doesn't report rejected addresses at send time in a form that can be
    # distinguished from other API errors. If SES rejects *any* recipient you'll get
    # an AnymailAPIError, and the message won't be sent to *all* recipients.

    # noinspection PyUnresolvedReferences
    def test_send_unparsable_response(self):
        """
        If the send succeeds, but result is unexpected format,
        should raise an API exception
        """
        response_content = {"wrong": "format"}
        self.set_mock_response(response_content)
        with self.assertRaisesMessage(
            AnymailAPIError, "parsing Amazon SES send result"
        ):
            self.message.send()
        self.assertIsNone(self.message.anymail_status.status)
        self.assertIsNone(self.message.anymail_status.message_id)
        self.assertEqual(self.message.anymail_status.recipients, {})
        self.assertEqual(self.message.anymail_status.esp_response, response_content)


@tag("amazon_ses")
class AmazonSESBackendConfigurationTests(AmazonSESBackendMockAPITestCase):
    """Test configuration options"""

    def test_boto_default_config(self):
        """By default, boto3 gets credentials from the environment or its config files

        See http://boto3.readthedocs.io/en/stable/guide/configuration.html
        """
        self.message.send()

        session_params = self.get_session_params()
        # no additional params passed to boto3.session.Session():
        self.assertEqual(session_params, {})

        client_params = self.get_client_params()
        # Ignore botocore.config.Config, which doesn't support ==
        config = client_params.pop("config")
        # no additional params passed to session.client('ses'):
        self.assertEqual(client_params, {})
        self.assertIn(
            f"django-anymail/{ANYMAIL_VERSION}-amazon-ses",
            config.user_agent_extra,
        )

    @override_settings(
        ANYMAIL={
            "AMAZON_SES_CLIENT_PARAMS": {
                # Example for testing; it's not a good idea to hardcode credentials in
                # your code. Safer: `os.getenv("MY_SPECIAL_AWS_KEY_ID")` etc.
                "aws_access_key_id": "test-access-key-id",
                "aws_secret_access_key": "test-secret-access-key",
                "region_name": "ap-northeast-1",
                # config can be given as dict of botocore.config.Config params
                "config": {
                    "read_timeout": 30,
                    "retries": {"max_attempts": 2},
                },
            }
        }
    )
    def test_client_params_in_setting(self):
        """
        The Anymail AMAZON_SES_CLIENT_PARAMS setting specifies
        boto3 session.client() params for Anymail
        """
        self.message.send()
        client_params = self.get_client_params()
        # Ignore botocore.config.Config, which doesn't support ==
        config = client_params.pop("config")
        self.assertEqual(
            client_params,
            {
                "aws_access_key_id": "test-access-key-id",
                "aws_secret_access_key": "test-secret-access-key",
                "region_name": "ap-northeast-1",
            },
        )
        self.assertEqual(config.read_timeout, 30)
        self.assertEqual(config.retries, {"max_attempts": 2})

    def test_client_params_in_connection_init(self):
        """
        You can also supply credentials specifically
        for a particular EmailBackend connection instance
        """
        from botocore.config import Config

        boto_config = Config(connect_timeout=30)
        conn = mail.get_connection(
            "anymail.backends.amazon_ses.EmailBackend",
            client_params={
                "aws_session_token": "test-session-token",
                "config": boto_config,
            },
        )
        conn.send_messages([self.message])

        client_params = self.get_client_params()
        # Ignore botocore.config.Config, which doesn't support ==
        config = client_params.pop("config")
        self.assertEqual(client_params, {"aws_session_token": "test-session-token"})
        self.assertEqual(config.connect_timeout, 30)

    @override_settings(
        ANYMAIL={"AMAZON_SES_SESSION_PARAMS": {"profile_name": "anymail-testing"}}
    )
    def test_session_params_in_setting(self):
        """
        The Anymail AMAZON_SES_SESSION_PARAMS setting
        specifies boto3.session.Session() params for Anymail
        """
        self.message.send()

        session_params = self.get_session_params()
        self.assertEqual(session_params, {"profile_name": "anymail-testing"})

        client_params = self.get_client_params()
        # Ignore botocore.config.Config, which doesn't support ==
        client_params.pop("config")
        # no additional params passed to session.client('ses'):
        self.assertEqual(client_params, {})

    @override_settings(
        ANYMAIL={"AMAZON_SES_CONFIGURATION_SET_NAME": "MyConfigurationSet"}
    )
    def test_config_set_setting(self):
        """You can supply a default ConfigurationSetName"""
        self.message.send()
        params = self.get_send_params()
        self.assertEqual(params["ConfigurationSetName"], "MyConfigurationSet")

        # override on individual message using esp_extra
        self.message.esp_extra = {"ConfigurationSetName": "CustomConfigurationSet"}
        self.message.send()
        params = self.get_send_params()
        self.assertEqual(params["ConfigurationSetName"], "CustomConfigurationSet")



# Anymail test utils
import re
import sys
import uuid
import warnings
from base64 import b64decode
from contextlib import contextmanager
from io import BytesIO, StringIO
from pathlib import Path
from unittest import TestCase
from unittest.util import safe_repr

import django.test.client


def decode_att(att):
    """Returns the original data from base64-encoded attachment content"""
    return b64decode(att.encode("ascii"))


def rfc822_unfold(text):
    # "Unfolding is accomplished by simply removing any CRLF that is immediately
    # followed by WSP" (WSP is space or tab, and per email.parser semantics, we allow
    # CRLF, CR, or LF endings)
    return re.sub(r"(\r\n|\r|\n)(?=[ \t])", "", text)


#
# Sample files for testing (in ./test_files subdir)
#

TEST_FILES_DIR = Path(__file__).parent.joinpath("test_files").resolve()

SAMPLE_IMAGE_FILENAME = "sample_image.png"
SAMPLE_EMAIL_FILENAME = "sample_email.txt"


def test_file_path(filename):
    """Returns path to a test file"""
    return TEST_FILES_DIR.joinpath(filename)


def test_file_content(filename):
    """Returns contents (bytes) of a test file"""
    return TEST_FILES_DIR.joinpath(filename).read_bytes()


def sample_image_path(filename=SAMPLE_IMAGE_FILENAME):
    """Returns path to an actual image file in the tests directory"""
    return test_file_path(filename)


def sample_image_content(filename=SAMPLE_IMAGE_FILENAME):
    """Returns contents of an actual image file from the tests directory"""
    return test_file_content(filename)


def sample_email_path(filename=SAMPLE_EMAIL_FILENAME):
    """Returns path to an email file (e.g., for forwarding as an attachment)"""
    return test_file_path(filename)


def sample_email_content(filename=SAMPLE_EMAIL_FILENAME):
    """
    Returns bytes contents of an email file (e.g., for forwarding as an attachment)
    """
    return test_file_content(filename)


#
# TestCase helpers
#


class AnymailTestMixin(TestCase):
    """Helpful additional methods for Anymail tests"""

    def assertDictMatches(self, expected, actual, msg=None):
        """
        Tests that dict `expected` is a subset of `actual`. (That expected
        is *in* actual. Note the args are in the same needle, haystack
        order as assertIn.)
        """
        # This is just assertDictContainsSubset, and the code is copied in from there.
        # (That was deprecated because apparently the arg order was confusing given the
        # name. The "matches" terminology is borrowed from several JS expect packages.)
        missing = []
        mismatched = []
        for key, value in expected.items():
            if key not in actual:
                missing.append(key)
            elif value != actual[key]:
                mismatched.append(
                    "%s, expected: %s, actual: %s"
                    % (safe_repr(key), safe_repr(value), safe_repr(actual[key]))
                )

        if not (missing or mismatched):
            return

        standardMsg = ""
        if missing:
            standardMsg = "Missing: %s" % ",".join(safe_repr(m) for m in missing)
        if mismatched:
            if standardMsg:
                standardMsg += "; "
            standardMsg += "Mismatched values: %s" % ",".join(mismatched)

        self.fail(self._formatMessage(msg, standardMsg))

    @contextmanager
    def assertDoesNotWarn(self, disallowed_warning=Warning):
        """Makes test error (rather than fail) if disallowed_warning occurs.

        Note: you probably want to be more specific than the default
        disallowed_warning=Warning, which errors for any warning
        (including DeprecationWarnings).
        """
        try:
            warnings.simplefilter("error", disallowed_warning)
            yield
        finally:
            warnings.resetwarnings()

    def assertEqualIgnoringHeaderFolding(self, first, second, msg=None):
        # Unfold (per RFC-8222) all text first and second, then compare result.
        # Useful for message/rfc822 attachment tests, where various Python email
        # versions handled folding slightly differently.
        # (Technically, this is unfolding both headers and (incorrectly) bodies,
        # but that doesn't really affect the tests.)
        if isinstance(first, bytes) and isinstance(second, bytes):
            first = first.decode("utf-8")
            second = second.decode("utf-8")
        first = rfc822_unfold(first)
        second = rfc822_unfold(second)
        self.assertEqual(first, second, msg)

    def assertUUIDIsValid(self, uuid_str, msg=None, version=4):
        """Assert the uuid_str evaluates to a valid UUID"""
        try:
            uuid.UUID(uuid_str, version=version)
        except (ValueError, AttributeError, TypeError):
            raise self.failureException(msg or "%r is not a valid UUID" % uuid_str)

    @contextmanager
    def assertPrints(self, expected, match="contain", msg=None):
        """Use as a context manager; checks that code writes `expected` to stdout.

        `match` can be "contain", "equal", "start", "end", or the name of any str
        method that takes one str argument and returns a boolean, or None to simply
        capture stdout without checking it. Default is "contain".

        Returns StringIO buffer; the output text is available as cm.getvalue().

        >>> with self.assertPrints("foo") as cm:
        ...     print("foo")
        >>> self.assertNotIn("bar", cm.getvalue())
        """
        matchfn = {
            "contain": "__contains__",
            "equal": "__eq__",
            "start": "startswith",
            "end": "endswith",
        }.get(match, match)
        old_stdout = sys.stdout
        buffer = StringIO()
        try:
            sys.stdout = buffer
            yield buffer
            if matchfn:
                actual = buffer.getvalue()
                bound_matchfn = getattr(actual, matchfn)
                if not bound_matchfn(expected):
                    raise self.failureException(
                        msg
                        or "Stdout {actual!r} does not {match} {expected!r}".format(
                            actual=actual, match=match, expected=expected
                        )
                    )
        finally:
            sys.stdout = old_stdout


class ClientWithCsrfChecks(django.test.Client):
    """Django test Client that enforces CSRF checks

    https://docs.djangoproject.com/en/stable/ref/csrf/#testing
    """

    def __init__(self, **defaults):
        super().__init__(enforce_csrf_checks=True, **defaults)


# dedent for bytestrs
# https://stackoverflow.com/a/39841195/647002
_whitespace_only_re = re.compile(b"^[ \t]+$", re.MULTILINE)
_leading_whitespace_re = re.compile(b"(^[ \t]*)(?:[^ \t\n])", re.MULTILINE)


def dedent_bytes(text):
    """textwrap.dedent, but for bytes"""
    # Look for the longest leading string of spaces and tabs common to
    # all lines.
    margin = None
    text = _whitespace_only_re.sub(b"", text)
    indents = _leading_whitespace_re.findall(text)
    for indent in indents:
        if margin is None:
            margin = indent

        # Current line more deeply indented than previous winner:
        # no change (previous winner is still on top).
        elif indent.startswith(margin):
            pass

        # Current line consistent with and no deeper than previous winner:
        # it's the new winner.
        elif margin.startswith(indent):
            margin = indent

        # Find the largest common whitespace between current line
        # and previous winner.
        else:
            for i, (x, y) in enumerate(zip(margin, indent)):
                if x != y:
                    margin = margin[:i]
                    break
            else:
                margin = margin[: len(indent)]

    if margin:
        text = re.sub(b"(?m)^" + margin, b"", text)
    return text


def make_fileobj(content, filename=None, content_type=None, encoding=None):
    """
    Returns a file-like object that can be used in Django test Client
    post data to simulate an uploaded file.
    """
    # The logic that unpacks this is in django.test.client.encode_file.
    if isinstance(content, str):
        content = content.encode(encoding or "utf-8")
    fileobj = BytesIO(content)
    if filename is not None:
        fileobj.name = filename
    if content_type is not None:
        fileobj.content_type = content_type
    return fileobj


def encode_multipart(boundary, data):
    """
    Version of :func:`django.test.client.encode_multipart` that allows
    empty filenames. (The original function substitutes the field's
    name if a file has an empty name.)
    """
    # For simplicity, encode with the original function, and then
    # replace any 'filename="<key>"' with 'filename=""'. This isn't
    # entirely robust, but is sufficient for testing use.
    encoded = django.test.client.encode_multipart(boundary, data)
    re_keys = r"|".join(re.escape(key) for key in data.keys())
    return re.sub(
        rb'filename="(%s)"' % re_keys.encode("ascii"), b'filename=""', encoded
    )



import json
from base64 import b64encode
from textwrap import dedent
from unittest.mock import ANY

from django.test import tag

from anymail.exceptions import AnymailConfigurationError
from anymail.inbound import AnymailInboundMessage
from anymail.signals import AnymailInboundEvent
from anymail.webhooks.postmark import PostmarkInboundWebhookView

from .utils import sample_email_content, sample_image_content, test_file_content
from .webhook_cases import WebhookTestCase


@tag("postmark")
class PostmarkInboundTestCase(WebhookTestCase):
    def test_inbound_basics(self):
        # without "Include raw email content in JSON payload"
        raw_event = {
            "FromFull": {
                "Email": "from+test@example.org",
                "Name": "Displayed From",
                "MailboxHash": "test",
            },
            "ToFull": [
                {
                    "Email": "test@inbound.example.com",
                    "Name": "Test Inbound",
                    "MailboxHash": "",
                },
                {"Email": "other@example.com", "Name": "", "MailboxHash": ""},
            ],
            "CcFull": [{"Email": "cc@example.com", "Name": "", "MailboxHash": ""}],
            "BccFull": [
                # Postmark provides Bcc if delivered-to address is not in To field:
                {
                    "Email": "test@inbound.example.com",
                    "Name": "",
                    "MailboxHash": "",
                }
            ],
            "OriginalRecipient": "test@inbound.example.com",
            "ReplyTo": "from+test@milter.example.org",
            "Subject": "Test subject",
            "MessageID": "22c74902-a0c1-4511-804f2-341342852c90",
            "Date": "Wed, 11 Oct 2017 18:31:04 -0700",
            "TextBody": "Test body plain",
            "HtmlBody": "<div>Test body html</div>",
            "StrippedTextReply": "stripped plaintext body",
            "Tag": "",
            "Headers": [
                {"Name": "Return-Path", "Value": "<envelope-from@example.org>"},
                {
                    "Name": "Received",
                    "Value": "from mail.example.org by inbound.postmarkapp.com ...",
                },
                {
                    "Name": "X-Spam-Checker-Version",
                    "Value": "SpamAssassin 3.4.0 (2014-02-07) on p-pm-smtp-inbound01b-aws-useast2b",  # NOQA: E501
                },
                {"Name": "X-Spam-Status", "Value": "No"},
                {"Name": "X-Spam-Score", "Value": "1.7"},
                {"Name": "X-Spam-Tests", "Value": "SPF_PASS"},
                {
                    "Name": "Received",
                    "Value": "by mail.example.org for <test@inbound.example.com> ...",
                },
                {
                    "Name": "Received",
                    "Value": "by 10.10.1.71 with HTTP; Wed, 11 Oct 2017 18:31:04 -0700 (PDT)",
                },
                {
                    "Name": "Return-Path",
                    "Value": "<fake-return-path@postmark-should-have-removed>",
                },
                {"Name": "MIME-Version", "Value": "1.0"},
                {"Name": "Message-ID", "Value": "<CAEPk3R+4Zr@mail.example.org>"},
            ],
        }

        response = self.client.post(
            "/anymail/postmark/inbound/",
            content_type="application/json",
            data=json.dumps(raw_event),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.inbound_handler,
            sender=PostmarkInboundWebhookView,
            event=ANY,
            esp_name="Postmark",
        )
        # AnymailInboundEvent
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailInboundEvent)
        self.assertEqual(event.event_type, "inbound")
        # Postmark doesn't provide inbound event timestamp:
        self.assertIsNone(event.timestamp)
        self.assertEqual(event.event_id, "22c74902-a0c1-4511-804f2-341342852c90")
        self.assertIsInstance(event.message, AnymailInboundMessage)
        self.assertEqual(event.esp_event, raw_event)

        # AnymailInboundMessage - convenience properties
        message = event.message

        self.assertEqual(message.from_email.display_name, "Displayed From")
        self.assertEqual(message.from_email.addr_spec, "from+test@example.org")
        self.assertEqual(
            [str(e) for e in message.to],
            ["Test Inbound <test@inbound.example.com>", "other@example.com"],
        )
        self.assertEqual([str(e) for e in message.cc], ["cc@example.com"])
        self.assertEqual([str(e) for e in message.bcc], ["test@inbound.example.com"])
        self.assertEqual(message.subject, "Test subject")
        self.assertEqual(message.date.isoformat(" "), "2017-10-11 18:31:04-07:00")
        self.assertEqual(message.text, "Test body plain")
        self.assertEqual(message.html, "<div>Test body html</div>")

        self.assertEqual(message.envelope_sender, "envelope-from@example.org")
        self.assertEqual(message.envelope_recipient, "test@inbound.example.com")
        self.assertEqual(message.stripped_text, "stripped plaintext body")
        # Postmark doesn't provide stripped html:
        self.assertIsNone(message.stripped_html)
        self.assertIs(message.spam_detected, False)
        self.assertEqual(message.spam_score, 1.7)

        # AnymailInboundMessage - other headers
        self.assertEqual(message["Message-ID"], "<CAEPk3R+4Zr@mail.example.org>")
        self.assertEqual(message["Reply-To"], "from+test@milter.example.org")
        self.assertEqual(
            message.get_all("Received"),
            [
                "from mail.example.org by inbound.postmarkapp.com ...",
                "by mail.example.org for <test@inbound.example.com> ...",
                "by 10.10.1.71 with HTTP; Wed, 11 Oct 2017 18:31:04 -0700 (PDT)",
            ],
        )

    def test_attachments(self):
        image_content = sample_image_content()
        email_content = sample_email_content()
        raw_event = {
            "Attachments": [
                {
                    "Name": "test.txt",
                    "Content": b64encode("test attachment".encode("utf-8")).decode(
                        "ascii"
                    ),
                    "ContentType": "text/plain",
                    "ContentLength": len("test attachment"),
                },
                {
                    "Name": "image.png",
                    "Content": b64encode(image_content).decode("ascii"),
                    "ContentType": "image/png",
                    "ContentID": "abc123",
                    "ContentLength": len(image_content),
                },
                {
                    "Name": "bounce.txt",
                    "Content": b64encode(email_content).decode("ascii"),
                    "ContentType": 'message/rfc822; charset="us-ascii"',
                    "ContentLength": len(email_content),
                },
            ]
        }

        response = self.client.post(
            "/anymail/postmark/inbound/",
            content_type="application/json",
            data=json.dumps(raw_event),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.inbound_handler,
            sender=PostmarkInboundWebhookView,
            event=ANY,
            esp_name="Postmark",
        )
        event = kwargs["event"]
        message = event.message
        attachments = message.attachments  # AnymailInboundMessage convenience accessor
        self.assertEqual(len(attachments), 2)
        self.assertEqual(attachments[0].get_filename(), "test.txt")
        self.assertEqual(attachments[0].get_content_type(), "text/plain")
        self.assertEqual(attachments[0].get_content_text(), "test attachment")
        self.assertEqual(attachments[1].get_content_type(), "message/rfc822")
        self.assertEqualIgnoringHeaderFolding(
            attachments[1].get_content_bytes(), email_content
        )

        inlines = message.content_id_map
        self.assertEqual(len(inlines), 1)
        inline = inlines["abc123"]
        self.assertEqual(inline.get_filename(), "image.png")
        self.assertEqual(inline.get_content_type(), "image/png")
        self.assertEqual(inline.get_content_bytes(), image_content)

    def test_inbound_with_raw_email(self):
        # With "Include raw email content in JSON payload"
        raw_event = {
            # (Postmark's "RawEmail" actually uses \n rather than \r\n)
            "RawEmail": dedent(
                """\
                Received: from mail.example.org by inbound.postmarkapp.com ...
                X-Spam-Checker-Version: SpamAssassin 3.4.0 (2014-02-07)
                \ton p-pm-smtp-inbound01b-aws-useast2b
                X-Spam-Status: No
                X-Spam-Score: 1.7
                X-Spam-Tests: SPF_PASS
                Received: by mail.example.org for <test@inbound.example.com> ...
                Received: by 10.10.1.71 with HTTP; Wed, 11 Oct 2017 18:31:04 -0700 (PDT)
                MIME-Version: 1.0
                Message-ID: <CAEPk3R+4Zr@mail.example.org>
                Return-Path: <fake-return-path@postmark-should-have-removed>
                From: "Displayed From" <from+test@example.org>
                To: Test Inbound <test@inbound.example.com>, other@example.com
                Cc: cc@example.com
                Reply-To: from+test@milter.example.org
                Subject: Test subject
                Date: Wed, 11 Oct 2017 18:31:04 -0700
                Content-Type: multipart/alternative; boundary="BoUnDaRy1"

                --BoUnDaRy1
                Content-Type: text/plain; charset="utf-8"
                Content-Transfer-Encoding: 7bit

                Test body plain
                --BoUnDaRy1
                Content-Type: multipart/related; boundary="bOuNdArY2"

                --bOuNdArY2
                Content-Type: text/html; charset="utf-8"
                Content-Transfer-Encoding: quoted-printable

                <div>Test body html</div>
                --bOuNdArY2
                Content-Type: text/plain
                Content-Transfer-Encoding: quoted-printable
                Content-Disposition: attachment; filename="attachment.txt"

                This is an attachment
                --bOuNdArY2--

                --BoUnDaRy1--
                """
            ),
            "BccFull": [
                # Postmark provides Bcc if delivered-to address is not in To field
                # (but not in RawEmail)
                {
                    "Email": "test@inbound.example.com",
                    "Name": "",
                    "MailboxHash": "",
                }
            ],
            "OriginalRecipient": "test@inbound.example.com",
            "MessageID": "22c74902-a0c1-4511-804f2-341342852c90",
            "StrippedTextReply": "stripped plaintext body",
            "Tag": "",
            "Headers": [
                {"Name": "Return-Path", "Value": "<envelope-from@example.org>"},
                # ... All the other headers would be here ...
                # This is a fake header (not in RawEmail) to make sure we only
                # add headers in one place:
                {
                    "Name": "X-No-Duplicates",
                    "Value": "headers only from RawEmail",
                },
            ],
            "Attachments": [
                # ... Real attachments would appear here. ...
                # This is a fake one (not in RawEmail) to make sure we only
                # add attachments in one place:
                {
                    "Name": "no-duplicates.txt",
                    "Content": b64encode("fake attachment".encode("utf-8")).decode(
                        "ascii"
                    ),
                    "ContentType": "text/plain",
                    "ContentLength": len("fake attachment"),
                },
            ],
        }

        response = self.client.post(
            "/anymail/postmark/inbound/",
            content_type="application/json",
            data=json.dumps(raw_event),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.inbound_handler,
            sender=PostmarkInboundWebhookView,
            event=ANY,
            esp_name="Postmark",
        )
        # AnymailInboundEvent
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailInboundEvent)
        self.assertEqual(event.event_type, "inbound")
        # Postmark doesn't provide inbound event timestamp:
        self.assertIsNone(event.timestamp)
        self.assertEqual(event.event_id, "22c74902-a0c1-4511-804f2-341342852c90")
        self.assertIsInstance(event.message, AnymailInboundMessage)
        self.assertEqual(event.esp_event, raw_event)

        # AnymailInboundMessage - convenience properties
        message = event.message

        self.assertEqual(message.from_email.display_name, "Displayed From")
        self.assertEqual(message.from_email.addr_spec, "from+test@example.org")
        self.assertEqual(
            [str(e) for e in message.to],
            ["Test Inbound <test@inbound.example.com>", "other@example.com"],
        )
        self.assertEqual([str(e) for e in message.cc], ["cc@example.com"])
        self.assertEqual([str(e) for e in message.bcc], ["test@inbound.example.com"])
        self.assertEqual(message.subject, "Test subject")
        self.assertEqual(message.date.isoformat(" "), "2017-10-11 18:31:04-07:00")
        self.assertEqual(message.text, "Test body plain")
        self.assertEqual(message.html, "<div>Test body html</div>")

        self.assertEqual(message.envelope_sender, "envelope-from@example.org")
        self.assertEqual(message.envelope_recipient, "test@inbound.example.com")
        self.assertEqual(message.stripped_text, "stripped plaintext body")
        # Postmark doesn't provide stripped html:
        self.assertIsNone(message.stripped_html)
        self.assertIs(message.spam_detected, False)
        self.assertEqual(message.spam_score, 1.7)

        # AnymailInboundMessage - other headers
        self.assertEqual(message["Message-ID"], "<CAEPk3R+4Zr@mail.example.org>")
        self.assertEqual(message["Reply-To"], "from+test@milter.example.org")
        self.assertEqual(
            message.get_all("Received"),
            [
                "from mail.example.org by inbound.postmarkapp.com ...",
                "by mail.example.org for <test@inbound.example.com> ...",
                "by 10.10.1.71 with HTTP; Wed, 11 Oct 2017 18:31:04 -0700 (PDT)",
            ],
        )

        # Attachments (from RawEmail only, not also from parsed):
        attachments = message.attachments
        self.assertEqual(len(attachments), 1)
        self.assertEqual(attachments[0].get_filename(), "attachment.txt")
        self.assertEqual(attachments[0].get_content_type(), "text/plain")
        self.assertEqual(attachments[0].get_content_text(), "This is an attachment")

        # Make sure we didn't load headers from both RawEmail and parsed:
        self.assertNotIn("X-No-Duplicates", message)

    def test_misconfigured_tracking(self):
        errmsg = (
            "You seem to have set Postmark's *Delivery* webhook"
            " to Anymail's Postmark *inbound* webhook URL."
        )
        with self.assertRaisesMessage(AnymailConfigurationError, errmsg):
            self.client.post(
                "/anymail/postmark/inbound/",
                content_type="application/json",
                data=json.dumps({"RecordType": "Delivery"}),
            )

    def test_check_payload(self):
        # Postmark's "Check" button in the inbound webhook dashboard posts a static
        # payload that doesn't match their docs or actual inbound message payloads.
        # (Its attachments have "Data" rather than "Content".)
        # They apparently have no plans to fix it, so make sure Anymail can handle it.
        for filename in [
            # Actual test payloads from 2023-05-05:
            "postmark-inbound-test-payload.json",
            "postmark-inbound-test-payload-with-raw.json",
        ]:
            with self.subTest(filename=filename):
                self.inbound_handler.reset_mock()  # (subTest doesn't setUp/tearDown)
                test_payload = test_file_content(filename)
                response = self.client.post(
                    "/anymail/postmark/inbound/",
                    content_type="application/json",
                    data=test_payload,
                )
                self.assertEqual(response.status_code, 200)
                self.assert_handler_called_once_with(
                    self.inbound_handler,
                    sender=PostmarkInboundWebhookView,
                    event=ANY,
                    esp_name="Postmark",
                )
                # Don't care about the actual test message contents here,
                # just want to make sure it parses and signals inbound without error.

    def test_spammy_address(self):
        # This long FromFull.Name caused an error in AnymailInboundMessage construction,
        # due to folding introduced in EmailAddress.formataddr()
        from_name = (
            "* * * 💲 Snag Your Free Gift! Click Here: "
            "https://spam.example.com/uploads/phish.php?123456 💲 * * *"
        )
        raw_event = {
            "MessageStream": "inbound",
            "FromFull": {
                "Email": "noreply@spam.example.com",
                "Name": from_name,
            },
        }
        response = self.client.post(
            "/anymail/postmark/inbound/",
            content_type="application/json",
            data=json.dumps(raw_event),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.inbound_handler,
            sender=PostmarkInboundWebhookView,
            event=ANY,
            esp_name="Postmark",
        )
        message = kwargs["event"].message
        self.assertEqual(message.from_email.display_name, from_name)



import os
import unittest
from email.utils import formataddr

from django.test import SimpleTestCase, override_settings, tag

from anymail.exceptions import AnymailAPIError
from anymail.message import AnymailMessage

from .utils import AnymailTestMixin, sample_image_path

ANYMAIL_TEST_AMAZON_SES_ACCESS_KEY_ID = os.getenv(
    "ANYMAIL_TEST_AMAZON_SES_ACCESS_KEY_ID"
)
ANYMAIL_TEST_AMAZON_SES_SECRET_ACCESS_KEY = os.getenv(
    "ANYMAIL_TEST_AMAZON_SES_SECRET_ACCESS_KEY"
)
ANYMAIL_TEST_AMAZON_SES_REGION_NAME = os.getenv(
    "ANYMAIL_TEST_AMAZON_SES_REGION_NAME", "us-east-1"
)
ANYMAIL_TEST_AMAZON_SES_DOMAIN = os.getenv("ANYMAIL_TEST_AMAZON_SES_DOMAIN")


@unittest.skipUnless(
    ANYMAIL_TEST_AMAZON_SES_ACCESS_KEY_ID
    and ANYMAIL_TEST_AMAZON_SES_SECRET_ACCESS_KEY
    and ANYMAIL_TEST_AMAZON_SES_DOMAIN,
    "Set ANYMAIL_TEST_AMAZON_SES_ACCESS_KEY_ID and"
    " ANYMAIL_TEST_AMAZON_SES_SECRET_ACCESS_KEY and ANYMAIL_TEST_AMAZON_SES_DOMAIN"
    " environment variables to run Amazon SES integration tests",
)
@override_settings(
    EMAIL_BACKEND="anymail.backends.amazon_ses.EmailBackend",
    ANYMAIL={
        "AMAZON_SES_CLIENT_PARAMS": {
            # This setting provides Anymail-specific AWS credentials to boto3.client(),
            # overriding any credentials in the environment or boto config. It's often
            # *not* the best approach. See the Anymail and boto3 docs for other options.
            "aws_access_key_id": ANYMAIL_TEST_AMAZON_SES_ACCESS_KEY_ID,
            "aws_secret_access_key": ANYMAIL_TEST_AMAZON_SES_SECRET_ACCESS_KEY,
            "region_name": ANYMAIL_TEST_AMAZON_SES_REGION_NAME,
            # Can supply any other boto3.client params,
            # including botocore.config.Config as dict
            "config": {"retries": {"max_attempts": 2}},
        },
        # actual config set in Anymail test account:
        "AMAZON_SES_CONFIGURATION_SET_NAME": "TestConfigurationSet",
    },
)
@tag("amazon_ses", "live")
class AmazonSESBackendIntegrationTests(AnymailTestMixin, SimpleTestCase):
    """Amazon SES API integration tests

    These tests run against the **live** Amazon SES API, using the environment
    variables `ANYMAIL_TEST_AMAZON_SES_ACCESS_KEY_ID` and
    `ANYMAIL_TEST_AMAZON_SES_SECRET_ACCESS_KEY` as AWS credentials.
    If those variables are not set, these tests won't run.

    (You can also set the environment variable `ANYMAIL_TEST_AMAZON_SES_REGION_NAME`
    to test SES using a region other than the default "us-east-1".)

    Amazon SES doesn't offer a test mode -- it tries to send everything you ask.
    To avoid stacking up a pile of undeliverable @example.com
    emails, the tests use Amazon's @simulator.amazonses.com addresses.
    https://docs.aws.amazon.com/ses/latest/DeveloperGuide/mailbox-simulator.html
    """

    def setUp(self):
        super().setUp()
        self.from_email = f"test@{ANYMAIL_TEST_AMAZON_SES_DOMAIN}"
        self.message = AnymailMessage(
            "Anymail Amazon SES integration test",
            "Text content",
            self.from_email,
            ["success@simulator.amazonses.com"],
        )
        self.message.attach_alternative("<p>HTML content</p>", "text/html")

    def test_simple_send(self):
        # Example of getting the Amazon SES send status and message id from the message
        sent_count = self.message.send()
        self.assertEqual(sent_count, 1)

        anymail_status = self.message.anymail_status
        sent_status = anymail_status.recipients[
            "success@simulator.amazonses.com"
        ].status
        message_id = anymail_status.recipients[
            "success@simulator.amazonses.com"
        ].message_id

        # Amazon SES always queues (or raises an error):
        self.assertEqual(sent_status, "queued")
        # Amazon SES message ids are groups of hex chars:
        self.assertRegex(message_id, r"[0-9a-f-]+")
        # set of all recipient statuses:
        self.assertEqual(anymail_status.status, {sent_status})
        self.assertEqual(anymail_status.message_id, message_id)

    def test_all_options(self):
        message = AnymailMessage(
            subject="Anymail Amazon SES all-options integration test",
            body="This is the text body",
            from_email=formataddr(("Test From, with comma", self.from_email)),
            to=[
                "success+to1@simulator.amazonses.com",
                "Recipient 2 <success+to2@simulator.amazonses.com>",
            ],
            cc=[
                "success+cc1@simulator.amazonses.com",
                "Copy 2 <success+cc2@simulator.amazonses.com>",
            ],
            bcc=[
                "success+bcc1@simulator.amazonses.com",
                "Blind Copy 2 <success+bcc2@simulator.amazonses.com>",
            ],
            reply_to=["reply1@example.com", "Reply 2 <reply2@example.com>"],
            headers={"X-Anymail-Test": "value"},
            metadata={"meta1": "simple_string", "meta2": 2},
            tags=["Re-engagement", "Cohort 12/2017"],
            envelope_sender=f"bounce-handler@{ANYMAIL_TEST_AMAZON_SES_DOMAIN}",
        )
        message.attach("attachment1.txt", "Here is some\ntext for you", "text/plain")
        message.attach("attachment2.csv", "ID,Name\n1,Amy Lina", "text/csv")
        cid = message.attach_inline_image_file(sample_image_path())
        message.attach_alternative(
            "<p><b>HTML:</b> with <a href='http://example.com'>link</a>"
            "and image: <img src='cid:%s'></div>" % cid,
            "text/html",
        )

        message.attach_alternative(
            "Amazon SES SendRawEmail actually supports multiple alternative parts",
            "text/x-note-for-email-geeks",
        )

        message.send()
        self.assertEqual(message.anymail_status.status, {"queued"})

    def test_stored_template(self):
        # Using a template created like this:
        # boto3.client('sesv2').create_email_template(
        #     TemplateName="TestTemplate",
        #     TemplateContent={
        #         "Subject": "Your order {{order}} shipped",
        #         "Html": "<h1>Dear {{name}}:</h1>"
        #                 "<p>Your order {{order}} shipped {{ship_date}}.</p>",
        #         "Text": "Dear {{name}}:\r\n"
        #                 "Your order {{order}} shipped {{ship_date}}."
        #     },
        # )
        message = AnymailMessage(
            template_id="TestTemplate",
            from_email=formataddr(("Test From", self.from_email)),
            to=[
                "First Recipient <success+to1@simulator.amazonses.com>",
                "success+to2@simulator.amazonses.com",
            ],
            merge_data={
                "success+to1@simulator.amazonses.com": {
                    "order": 12345,
                    "name": "Test Recipient",
                },
                "success+to2@simulator.amazonses.com": {"order": 6789},
            },
            merge_global_data={"name": "Customer", "ship_date": "today"},  # default
            headers={
                "List-Unsubscribe-Post": "List-Unsubscribe=One-Click",
            },
            merge_headers={
                "success+to1@simulator.amazonses.com": {
                    "List-Unsubscribe": "<https://example.com/unsubscribe/to1>"
                },
                "success+to2@simulator.amazonses.com": {
                    "List-Unsubscribe": "<https://example.com/unsubscribe/to2>"
                },
            },
            tags=["Live integration test", "Template send"],
            metadata={"test": "data"},
            merge_metadata={"success+to2@simulator.amazonses.com": {"user-id": "2"}},
        )
        message.send()
        recipient_status = message.anymail_status.recipients
        self.assertEqual(
            recipient_status["success+to1@simulator.amazonses.com"].status, "queued"
        )
        self.assertRegex(
            recipient_status["success+to1@simulator.amazonses.com"].message_id,
            r"[0-9a-f-]+",
        )
        self.assertEqual(
            recipient_status["success+to2@simulator.amazonses.com"].status, "queued"
        )
        self.assertRegex(
            recipient_status["success+to2@simulator.amazonses.com"].message_id,
            r"[0-9a-f-]+",
        )

    @override_settings(
        ANYMAIL={
            "AMAZON_SES_CLIENT_PARAMS": {
                "aws_access_key_id": "test-invalid-access-key-id",
                "aws_secret_access_key": "test-invalid-secret-access-key",
                "region_name": ANYMAIL_TEST_AMAZON_SES_REGION_NAME,
            }
        }
    )
    def test_invalid_aws_credentials(self):
        # Make sure the exception message includes AWS's response:
        with self.assertRaisesMessage(
            AnymailAPIError, "The security token included in the request is invalid"
        ):
            self.message.send()
